"""
Standardized class and functions to test instruments for pysat libraries.  Not
directly called by pytest, but imported as part of test_instruments.py.  Can
be imported directly for external instrument libraries of pysat instruments.
"""
import datetime as dt
from importlib import import_module
import warnings

import pytest

import pysat


def initialize_test_inst_and_date(inst_dict):
    """Initializes the instrument object to test and date

    Paramters
    ---------
    inst_dict : dict
        Dictionary containing specific instrument info, generated by
        generate_instrument_list

    Returns
    -------
    test_inst : pysat.Instrument
        instrument object to be tested
    date : dt.datetime
        test date from module

    """

    test_inst = pysat.Instrument(inst_module=inst_dict['inst_module'],
                                 tag=inst_dict['tag'],
                                 inst_id=inst_dict['inst_id'],
                                 temporary_file_list=True,
                                 update_files=True)
    test_dates = inst_dict['inst_module']._test_dates
    date = test_dates[inst_dict['inst_id']][inst_dict['tag']]
    return test_inst, date


class InstTestClass():
    """Provides standardized tests for pysat instrument libraries.

    Note: Not diretly run by pytest, but inherited through test_instruments.py
    """
    module_attrs = ['platform', 'name', 'tags', 'inst_ids',
                    'load', 'list_files', 'download']
    inst_attrs = ['tag', 'inst_id', 'acknowledgements', 'references',
                  'inst_module']
    inst_callable = ['load', 'list_files', 'download', 'clean', 'default']
    attr_types = {'platform': str, 'name': str, 'tags': dict,
                  'inst_ids': dict, 'tag': str, 'inst_id': str,
                  'acknowledgements': str, 'references': str}

    def assert_hasattr(self, obj, attr_name):
        """ Nice assertion statement for `assert hasattr(obj, attr_name)`
        """
        estr = "Object {:} missing attribute {:}".format(obj.__repr__(),
                                                         attr_name)
        assert hasattr(obj, attr_name), estr

    def assert_isinstance(self, obj, obj_type):
        """ Nice assertion statement for `assert isinstance(obj, obj_type)`
        """
        estr = "Object {:} is type {:}, but should be type {:}".format(
            obj.__repr__(), type(obj), obj_type)
        assert isinstance(obj, obj_type), estr

    @pytest.mark.all_inst
    def test_modules_standard(self, inst_name):
        """Checks that modules are importable and have standard properties.
        """
        # ensure that each module is at minimum importable
        module = import_module(''.join(('.', inst_name)),
                               package=self.inst_loc.__name__)
        # Check for presence of basic instrument module attributes
        for mattr in self.module_attrs:
            self.assert_hasattr(module, mattr)
            if mattr in self.attr_types.keys():
                self.assert_isinstance(getattr(module, mattr),
                                       self.attr_types[mattr])

        # Check for presence of required instrument attributes
        for inst_id in module.inst_ids.keys():
            for tag in module.inst_ids[inst_id]:
                inst = pysat.Instrument(inst_module=module, tag=tag,
                                        inst_id=inst_id)

                # Test to see that the class parameters were passed in
                self.assert_isinstance(inst, pysat.Instrument)
                assert inst.platform == module.platform
                assert inst.name == module.name
                assert inst.inst_id == inst_id
                assert inst.tag == tag

                # Test the required class attributes
                for iattr in self.inst_attrs:
                    self.assert_hasattr(inst, iattr)
                    if iattr in self.attr_types:
                        self.assert_isinstance(getattr(inst, iattr),
                                               self.attr_types[iattr])

    @pytest.mark.all_inst
    def test_standard_function_presence(self, inst_name):
        """Check if each function is callable and all required functions exist
        """
        module = import_module(''.join(('.', inst_name)),
                               package=self.inst_loc.__name__)

        # Test for presence of all standard module functions
        for mcall in self.inst_callable:
            if hasattr(module, mcall):
                # If present, must be a callable function
                assert callable(getattr(module, mcall))
            else:
                # If absent, must not be a required function
                assert mcall not in self.module_attrs

    @pytest.mark.all_inst
    def test_instrument_test_dates(self, inst_name):
        """Check that module has structured test dates correctly."""
        module = import_module(''.join(('.', inst_name)),
                               package=self.inst_loc.__name__)
        info = module._test_dates
        for inst_id in info.keys():
            for tag in info[inst_id].keys():
                self.assert_isinstance(info[inst_id][tag], dt.datetime)

    @pytest.mark.first
    @pytest.mark.download
    def test_download(self, inst_dict):
        """Check that instruments are downloadable."""

        test_inst, date = initialize_test_inst_and_date(inst_dict)

        # check for username
        dl_dict = inst_dict['user_info'] if 'user_info' in \
            inst_dict.keys() else {}
        test_inst.download(date, date, **dl_dict)
        assert len(test_inst.files.files) > 0

    @pytest.mark.second
    @pytest.mark.download
    @pytest.mark.parametrize("clean_level", ['none', 'dirty', 'dusty', 'clean'])
    def test_load(self, clean_level, inst_dict):
        """Check that instruments load at each cleaning level."""

        test_inst, date = initialize_test_inst_and_date(inst_dict)
        if len(test_inst.files.files) > 0:
            # Set Clean Level
            test_inst.clean_level = clean_level
            target = 'Fake Data to be cleared'
            test_inst.data = [target]
            try:
                test_inst.load(date=date)
            except ValueError as verr:
                # Check if instrument is failing due to strict time flag
                if str(verr).find('Loaded data') > 0:
                    test_inst.strict_time_flag = False
                    with warnings.catch_warnings(record=True) as war:
                        test_inst.load(date=date)
                    assert len(war) >= 1
                    categories = [war[j].category for j in range(0, len(war))]
                    assert UserWarning in categories
                else:
                    # If error message does not match, raise error anyway
                    raise(verr)

            # Make sure fake data is cleared
            assert target not in test_inst.data
            # If cleaning not used, something should be in the file
            # Not used for clean levels since cleaning may remove all data
            if clean_level == "none":
                assert not test_inst.empty
        else:
            pytest.skip("Download data not available")

    @pytest.mark.download
    def test_remote_file_list(self, inst_dict):
        """Check if optional list_remote_files routine exists and is callable.
        """
        test_inst, date = initialize_test_inst_and_date(inst_dict)
        name = '_'.join((test_inst.platform, test_inst.name))

        if hasattr(getattr(self.inst_loc, name), 'list_remote_files'):
            assert callable(test_inst.remote_file_list)
            # check for username
            dl_dict = inst_dict['user_info'] if 'user_info' in \
                inst_dict.keys() else {}
            files = test_inst.remote_file_list(start=date, stop=date, **dl_dict)
            # If test date is correctly chosen, files should exist
            assert len(files) > 0
        else:
            pytest.skip("remote_file_list not available")

    @pytest.mark.no_download
    def test_download_warning(self, inst_dict):
        """Check that instruments without download support have a warning."""
        test_inst, date = initialize_test_inst_and_date(inst_dict)

        with warnings.catch_warnings(record=True) as war:
            test_inst.download(date, date)

        assert len(war) >= 1
        categories = [war[j].category for j in range(0, len(war))]
        assert UserWarning in categories



"""
pysat.model_utils - interface for modeled observations
======================================================

Main Features
-------------
- Comparison of models and measured data.
- Matching of instruments to modelled data parameters.
- Extraction of instrument-aligned data from a modelled data set.
"""

from __future__ import print_function
from __future__ import absolute_import

import datetime as dt
import numpy as np
import pandas as pds
import warnings

from pysat import logger


def satellite_view_through_model(sat, tie, scoords, tlabels):
    """Interpolate model values onto satellite orbital path.

    .. deprecated:: 2.2.0
      `satellite_view_through_model` will be removed in pysat 3.0.0, it will
      be added to pysatModels

    Parameters
    ----------
    sat : pysat.Instrument object
        Instrument object with some form of coordinates
    tie : ucar_tiegcm object
        Model run loaded as tie_gcm object
    scoords : string or list of strings
        Variable names reflecting coordinates in sat to interpolate model onto
    tlabels : string or list of strings
        Variable names from model to interpolate onto sat locations
    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatModelUtils instead:"
                            "https://github.com/pysat/pysatModelUtils"]),
                  DeprecationWarning, stacklevel=2)

    # tiegcm is in pressure levels, need in altitude, but on regular
    # grid
    import scipy.interpolate as interpolate

    # create input array using satellite time/position
    if isinstance(scoords, str):
        scoords = [scoords]
    coords = [sat[coord] for coord in scoords]
    coords.insert(0, sat.index.values.astype(int))
    sat_pts = [inp for inp in zip(*coords)]

    interp = {}
    if isinstance(tlabels, str):
        tlabels = [tlabels]
    for label in tlabels:
        points = [tie.data.coords[dim].values if dim != 'time' else
                  tie.data.coords[dim].values.astype(int)
                  for dim in tie[label].dims]
        interp[label] = interpolate.RegularGridInterpolator(points,
                                                            tie[label].values,
                                                            bounds_error=False,
                                                            fill_value=None)
        sat[''.join(('model_', label))] = interp[label](sat_pts)


def compare_model_and_inst(pairs=None, inst_name=[], mod_name=[],
                           methods=['all']):
    """Compare modelled and measured data

    .. deprecated:: 2.2.0
      `satellite_view_through_model` will be removed in pysat 3.0.0, it will
      be added to pysatModels

    Parameters
    ------------
    pairs : xarray.Dataset instance
        Dataset containing only the desired observation-model data pairs
    inst_name : list of strings
        ordered list of instrument measurements to compare to modelled data
    mod_name : list of strings
        ordered list of modelled data to compare to instrument measurements
    methods : list of strings
        statistics to calculate.  See Notes for accecpted inputs

    Returns
    ----------
    stat_dict : dict of dicts
        Dictionary where the first layer of keys denotes the instrument data
        name and the second layer provides the desired statistics
    data_units : dict
        Dictionary containing the units for the data

    Notes
    -----
    Statistics are calculated using PyForecastTools (imported as verify).
    See notes there for more details.

    all - all statistics
    all_bias - bias, meanPercentageError, medianLogAccuracy,
               symmetricSignedBias
    accuracy - returns dict with mean squared error, root mean squared error,
               mean absolute error, and median absolute error
    scaledAccuracy - returns dict with normaled root mean squared error, mean
                     absolute scaled error, mean absolute percentage error,
                     median absolute percentage error, median symmetric
                     accuracy
    bias - scale-dependent bias as measured by the mean error
    meanPercentageError - mean percentage error
    medianLogAccuracy - median of the log accuracy ratio
    symmetricSignedBias - Symmetric signed bias, as a percentage
    meanSquaredError - mean squared error
    RMSE - root mean squared error
    meanAbsError - mean absolute error
    medAbsError - median absolute error
    nRMSE - normaized root mean squared error
    scaledError - scaled error (see PyForecastTools for references)
    MASE - mean absolute scaled error
    forecastError - forecast error (see PyForecastTools for references)
    percError - percentage error
    absPercError - absolute percentage error
    logAccuracy - log accuracy ratio
    medSymAccuracy - Scaled measure of accuracy
    meanAPE - mean absolute percentage error

    """
    import verify  # PyForecastTools
    from pysat import utils

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatModelUtils instead:"
                            "https://github.com/pysat/pysatModelUtils"]),
                  DeprecationWarning, stacklevel=2)

    method_rout = {"bias": verify.bias, "accuracy": verify.accuracy,
                   "meanPercentageError": verify.meanPercentageError,
                   "medianLogAccuracy": verify.medianLogAccuracy,
                   "symmetricSignedBias": verify.symmetricSignedBias,
                   "meanSquaredError": verify.meanSquaredError,
                   "RMSE": verify.RMSE, "meanAbsError": verify.meanAbsError,
                   "medAbsError": verify.medAbsError, "MASE": verify.MASE,
                   "scaledAccuracy": verify.scaledAccuracy,
                   "nRMSE": verify.nRMSE, "scaledError": verify.scaledError,
                   "forecastError": verify.forecastError,
                   "percError": verify.percError, "meanAPE": verify.meanAPE,
                   "absPercError": verify.absPercError,
                   "logAccuracy": verify.logAccuracy,
                   "medSymAccuracy": verify.medSymAccuracy}

    replace_keys = {'MSE': 'meanSquaredError', 'MAE': 'meanAbsError',
                    'MdAE': 'medAbsError', 'MAPE': 'meanAPE',
                    'MdSymAcc': 'medSymAccuracy'}

    # Grouped methods for things that don't have convenience functions
    grouped_methods = {"all_bias": ["bias", "meanPercentageError",
                                    "medianLogAccuracy",
                                    "symmetricSignedBias"],
                       "all": list(method_rout.keys())}

    # Replace any group method keys with the grouped methods
    for gg in [(i, mm) for i, mm in enumerate(methods)
               if mm in list(grouped_methods.keys())]:
        # Extend the methods list to include all the grouped methods
        methods.extend(grouped_methods[gg[1]])
        # Remove the grouped method key
        methods.pop(gg[0])

    # Ensure there are no duplicate methods
    methods = list(set(methods))

    # Test the input
    if pairs is None:
        raise ValueError('must provide Dataset of paired observations')

    if len(inst_name) != len(mod_name):
        raise ValueError('must provide equal number of instrument and model ' +
                         'data names for comparison')

    if not np.all([iname in pairs.data_vars.keys() for iname in inst_name]):
        raise ValueError('unknown instrument data value supplied')

    if not np.all([iname in pairs.data_vars.keys() for iname in mod_name]):
        raise ValueError('unknown model data value supplied')

    if not np.all([mm in list(method_rout.keys()) for mm in methods]):
        known_methods = list(method_rout.keys())
        known_methods.extend(list(grouped_methods.keys()))
        unknown_methods = [mm for mm in methods
                           if mm not in list(method_rout.keys())]
        raise ValueError('unknown statistical method(s) requested:\n' +
                         '{:}\nuse only:\n{:}'.format(unknown_methods,
                                                      known_methods))

    # Initialize the output
    stat_dict = {iname: dict() for iname in inst_name}
    data_units = {iname: pairs.data_vars[iname].units for iname in inst_name}

    # Cycle through all of the data types
    for i, iname in enumerate(inst_name):
        # Determine whether the model data needs to be scaled
        iscale = utils.scale_units(pairs.data_vars[iname].units,
                                   pairs.data_vars[mod_name[i]].units)
        mod_scaled = pairs.data_vars[mod_name[i]].values.flatten() * iscale

        # Flatten both data sets, since accuracy routines require 1D arrays
        inst_dat = pairs.data_vars[iname].values.flatten()

        # Ensure no NaN are used in statistics
        inum = np.where(np.isfinite(mod_scaled) & np.isfinite(inst_dat))[0]


        if inum.shape[0] < 2:
            # Not all data types can use all statistics.  Print warnings
            # instead of stopping processing.  Only valid statistics
            # will be included in output
            logger.info("{:s} can't calculate stats for {:d} finite samples".format( \
                                                        iname, inum.shape[0]))
            stat_dict
        else:
            # Calculate all of the desired statistics
            for mm in methods:
                try:
                    stat_dict[iname][mm] = method_rout[mm](mod_scaled[inum],
                                                           inst_dat[inum])

                    # Convenience functions add layers to the output, remove
                    # these layers
                    if hasattr(stat_dict[iname][mm], "keys"):
                        for nn in stat_dict[iname][mm].keys():
                            new = replace_keys[nn] if nn in replace_keys.keys()\
                                else nn
                            stat_dict[iname][new] = stat_dict[iname][mm][nn]
                        del stat_dict[iname][mm]
                except ValueError as verr:
                    # Not all data types can use all statistics.  Print warnings
                    # instead of stopping processing.  Only valid statistics
                    # will be included in output
                    logger.warning("{:s} can't use {:s}: {:}".format(iname, mm, verr))
                except NotImplementedError:
                    # Not all data types can use all statistics.  Print warnings
                    # instead of stopping processing.  Only valid statistics
                    # will be included in output
                    logger.warning("{:s} can't implement {:s}".format(iname, mm))

    return stat_dict, data_units


def collect_inst_model_pairs(start=None, stop=None, tinc=None, inst=None,
                             user=None, password=None, model_files=None,
                             model_load_rout=None, inst_lon_name=None,
                             mod_lon_name=None, inst_name=[], mod_name=[],
                             mod_datetime_name=None, mod_time_name=None,
                             mod_units=[], sel_name=None, method='linear',
                             model_label='model', inst_clean_rout=None,
                             comp_clean='clean'):
    """Pair instrument and model data, applying data cleaning after finding the
    times and locations where the instrument and model align

    .. deprecated:: 2.2.0
      `collect_inst_model_pairs` will be removed in pysat 3.0.0, it will
      be added to pysatModels

    Parameters
    ----------
    start : dt.datetime
        Starting datetime
    stop : dt.datetime
        Ending datetime
    tinc : dt.timedelta
        Time incriment for model files
    inst : pysat.Instrument instance
        instrument object for which modelled data will be extracted
    user : string
        User name (needed for some data downloads)
    password : string
        Password (needed for some data downloads)
    model_files : string
        string format that will construct the desired model filename from a
        datetime object
    model_load_rout : routine
        Routine to load model data into an xarray using filename and datetime
        as input
    inst_lon_name : string
        variable name for instrument longitude
    mod_lon_name : string
        variable name for model longitude
    inst_name : list of strings
        list of names of the data series to use for determing instrument
        location
    mod_name : list of strings
        list of names of the data series to use for determing model locations
        in the same order as inst_name.  These must make up a regular grid.
    mod_datetime_name : string
        Name of the data series in the model Dataset containing datetime info
    mod_time_name : string
        Name of the time coordinate in the model Dataset
    mod_units : list of strings
        units for each of the mod_name location attributes.  Currently
        supports: rad/radian(s), deg/degree(s), h/hr(s)/hour(s), m, km, and cm
    sel_name : list of strings or NoneType
        list of names of modelled data indices to append to instrument object,
        or None to append all modelled data (default=None)
    method : string
        Interpolation method.  Supported are 'linear', 'nearest', and
        'splinef2d'.  The last is only supported for 2D data and is not
        recommended here.  (default='linear')
    model_label : string
        name of model, used to identify interpolated data values in instrument
        (default="model")
    inst_clean_rout : routine
        Routine to clean the instrument data
    comp_clean : string
        Clean level for the comparison data ('clean', 'dusty', 'dirty', 'none')
        (default='clean')

    Returns
    -------
    matched_inst : pysat.Instrument instance
        instrument object and paired modelled data

    """
    from os import path
    import pysat

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatModelUtils instead:"
                            "https://github.com/pysat/pysatModelUtils"]),
                  DeprecationWarning, stacklevel=2)

    matched_inst = None

    # Test the input
    if start is None or stop is None:
        raise ValueError('Must provide start and end time for comparison')

    if inst is None:
        raise ValueError('Must provide a pysat instrument object')

    if model_files is None:
        raise ValueError('Must provide list of modelled data')

    if model_load_rout is None:
        raise ValueError('Need routine to load modelled data')

    if mod_datetime_name is None:
        raise ValueError('Need time coordinate name for model datasets')

    if mod_time_name is None:
        raise ValueError('Need time coordinate name for model datasets')

    if len(inst_name) == 0:
        estr = 'Must provide instrument location attribute names as a list'
        raise ValueError(estr)

    if len(inst_name) != len(mod_name):
        estr = 'Must provide the same number of instrument and model '
        estr += 'location attribute names as a list'
        raise ValueError(estr)

    if len(mod_name) != len(mod_units):
        raise ValueError('Must provide units for each model location ' +
                         'attribute')

    if inst_clean_rout is None:
        raise ValueError('Need routine to clean the instrument data')

    # Download the instrument data, if needed
    # Could use some improvement, for not re-downloading times that you already
    # have
    if (stop - start).days != len(inst.files[start:stop]):
        inst.download(start=start, stop=stop, user=user, password=password)

    # Cycle through the times, loading the model and instrument data as needed
    istart = start
    while start < stop:
        mod_file = start.strftime(model_files)

        if path.isfile(mod_file):
            try:
                mdata = model_load_rout(mod_file, start)
                lon_high = float(mdata.coords[mod_lon_name].max())
                lon_low = float(mdata.coords[mod_lon_name].min())
            except Exception as err:
                print("unable to load {:s}: {:}".format(mod_file, err))
                mdata = None
        else:
            mdata = None

        if mdata is not None:
            # Load the instrument data, if needed
            if inst.empty or inst.index[-1] < istart:
                inst.custom.add(pysat.utils.coords.update_longitude, 'modify',
                                low=lon_low, lon_name=inst_lon_name,
                                high=lon_high)
                inst.load(date=istart)

            if not inst.empty and inst.index[0] >= istart:
                added_names = extract_modelled_observations(inst=inst, \
                                        model=mdata, \
                                        inst_name=inst_name, \
                                        mod_name=mod_name, \
                                        mod_datetime_name=mod_datetime_name, \
                                        mod_time_name=mod_time_name, \
                                        mod_units=mod_units, \
                                        sel_name=sel_name, \
                                        method=method, \
                                        model_label=model_label)

                if len(added_names) > 0:
                    # Clean the instrument data
                    inst.clean_level = comp_clean
                    inst_clean_rout(inst)

                    im = list()
                    for aname in added_names:
                        # Determine the number of good points
                        if inst.pandas_format:
                            imnew = np.where(np.isfinite(inst[aname]))
                        else:
                            imnew = np.where(np.isfinite(inst[aname].values))

                        # Some data types are higher dimensions than others,
                        # make sure we end up choosing a high dimension one
                        # so that we don't accidently throw away paired data
                        if len(im) == 0 or len(im[0]) < len(imnew[0]):
                            im = imnew

                    # If the data is 1D, save it as a list instead of a tuple
                    if len(im) == 1:
                        im = im[0]
                    else:
                        im = {kk: np.unique(im[i])
                              for i, kk in enumerate(inst.data.coords.keys())}

                    # Save the clean, matched data
                    if matched_inst is None:
                        matched_inst = pysat.Instrument
                        matched_inst.meta = inst.meta
                        matched_inst.data = inst[im]
                    else:
                        idata = inst[im]
                        matched_inst.data = \
                            inst.concat_data([matched_inst.data, idata])

                    # Reset the clean flag
                    inst.clean_level = 'none'

        # Cycle the times
        if tinc.total_seconds() <= 86400.0:
            start += tinc
            if start + tinc > istart + dt.timedelta(days=1):
                istart += dt.timedelta(days=1)
        else:
            if start + tinc >= istart + dt.timedelta(days=1):
                istart += dt.timedelta(days=1)
            if istart >= start + tinc:
                start += tinc

    # Recast as xarray and add units
    if matched_inst is not None:
        if inst.pandas_format:
            matched_inst.data = matched_inst.data.to_xarray()
        for im in inst.meta.data.units.keys():
            if im in matched_inst.data.data_vars.keys():
                matched_inst.data.data_vars[im].attrs['units'] = \
                    inst.meta.data.units[im]

    return matched_inst


def extract_modelled_observations(inst=None, model=None, inst_name=[],
                                  mod_name=[], mod_datetime_name=None,
                                  mod_time_name=None, mod_units=[],
                                  sel_name=None, method='linear',
                                  model_label='model'):
    """Extracts instrument-aligned data from a modelled data set

    .. deprecated:: 2.2.0
      `extract_modelled_observations` will be removed in pysat 3.0.0, it will
      be added to pysatModels

    Parameters
    ----------
    inst : pysat.Instrument instance
        instrument object for which modelled data will be extracted
    model : xarray Dataset
        modelled data set
    inst_name : list of strings
        list of names of the data series to use for determing instrument
        location
    mod_name : list of strings
        list of names of the data series to use for determing model locations
        in the same order as inst_name.  These must make up a regular grid.
    mod_datetime_name : string
        Name of the data series in the model Dataset containing datetime info
    mod_time_name : string
        Name of the time coordinate in the model Dataset
    mod_units : list of strings
        units for each of the mod_name location attributes.  Currently
        supports: rad/radian(s), deg/degree(s), h/hr(s)/hour(s), m, km, and cm
    sel_name : list of strings or NoneType
        list of names of modelled data indices to append to instrument object,
        or None to append all modelled data (default=None)
    method : string
        Interpolation method.  Supported are 'linear', 'nearest', and
        'splinef2d'.  The last is only supported for 2D data and is not
        recommended here.  (default='linear')
    model_label : string
        name of model, used to identify interpolated data values in instrument
        (default="model")

    Returns
    -------
    added_names : list of strings
        list of names of modelled data added to the instrument

    Notes
    --------
    For best results, select clean instrument data after alignment with model

    """
    from scipy import interpolate
    from pysat import utils

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatModelUtils instead:"
                            "https://github.com/pysat/pysatModelUtils"]),
                  DeprecationWarning, stacklevel=2)

    # Test input
    if inst is None:
        raise ValueError('Must provide a pysat instrument object')

    if model is None:
        raise ValueError('Must provide modelled data')

    if mod_datetime_name is None:
        raise ValueError('Need datetime key for model datasets')

    if mod_time_name is None:
        raise ValueError('Need time coordinate name for model datasets')

    if len(inst_name) == 0:
        estr = 'Must provide instrument location attribute names as a list'
        raise ValueError(estr)

    if len(inst_name) != len(mod_name):
        estr = 'Must provide the same number of instrument and model '
        estr += 'location attribute names as a list'
        raise ValueError(estr)

    if len(mod_name) != len(mod_units):
        raise ValueError('Must provide units for each model location ' +
                         'attribute')

    inst_scale = np.ones(shape=len(inst_name), dtype=float)
    for i, ii in enumerate(inst_name):
        if ii not in list(inst.data.keys()):
            raise ValueError('Unknown instrument location index ' +
                             '{:}'.format(ii))
        inst_scale[i] = utils.scale_units(mod_units[i],
                                          inst.meta.data.units[ii])

    # Determine which data to interpolate and initialize the interpolated
    # output
    if sel_name is None:
        sel_name = list(model.data_vars.keys())

    for mi in mod_name:
        if mi in sel_name:
            sel_name.pop(sel_name.index(mi))

    # Determine the model time resolution
    tm_sec = (np.array(model.data_vars[mod_datetime_name][1:]) -
              np.array(model.data_vars[mod_datetime_name][:-1])).min()
    tm_sec /= np.timedelta64(1, 's')
    ti_sec = (inst.index[1:] - inst.index[:-1]).min().total_seconds()
    min_del = tm_sec if tm_sec < ti_sec else ti_sec

    # Determine which instrument observations are within the model time
    # resolution of a model run
    mind = list()
    iind = list()
    for i, tt in enumerate(np.array(model.data_vars[mod_datetime_name])):
        del_sec = abs(tt - inst.index).total_seconds()
        if del_sec.min() <= min_del:
            iind.append(del_sec.argmin())
            mind.append(i)

    # Determine the model coordinates closest to the satellite track
    interp_data = dict()
    interp_shape = inst.index.shape if inst.pandas_format else \
        inst.data.data_vars.items()[0][1].shape
    inst_coord = {kk: getattr(inst.data, inst_name[i]).values * inst_scale[i]
                  for i, kk in enumerate(mod_name)}
    for i, ii in enumerate(iind):
        # Cycle through each model data type, since it may not depend on
        # all the dimensions
        for mdat in sel_name:
            # Determine the dimension values
            dims = list(model.data_vars[mdat].dims)
            ndim = model.data_vars[mdat].data.shape
            indices = {mod_time_name: mind[i]}

            # Construct the data needed for interpolation
            values = model[indices][mdat].data
            points = [model.coords[kk].data for kk in dims if kk in mod_name]
            get_coords = True if len(points) > 0 else False
            idims = 0

            while get_coords:
                if inst.pandas_format:
                    # This data iterates only by time
                    xout = ii
                    xi = [inst_coord[kk][ii] for kk in dims if kk in mod_name]
                    get_coords = False
                else:
                    # This data may have additional dimensions
                    if idims == 0:
                        # Determine the number of dimensions
                        idims = len(inst.data.coords)
                        idim_names = inst.data.coords.keys()[1:]

                        # Find relevent dimensions for cycling and slicing
                        ind_dims = [k for k, kk in enumerate(inst_name)
                                    if kk in idim_names]
                        imod_dims = [k for k in ind_dims
                                     if mod_name[k] in dims]
                        ind_dims = [inst.data.coords.keys().index(inst_name[k])
                                    for k in imod_dims]

                        # Set the number of cycles
                        icycles = 0
                        ncycles = sum([len(inst.data.coords[inst_name[k]])
                                       for k in imod_dims])
                        cinds = np.zeros(shape=len(imod_dims), dtype=int)

                    # Get the instrument coordinate for this cycle
                    if icycles < ncycles or icycles == 0:
                        ss = [ii if k == 0 else 0 for k in range(idims)]
                        se = [ii + 1 if k == 0 else
                              len(inst.data.coords[idim_names[k-1]])
                              for k in range(idims)]
                        xout = [cinds[ind_dims.index(k)] if k in ind_dims
                                else slice(ss[k], se[k]) for k in range(idims)]
                        xind = [cinds[ind_dims.index(k)] if k in ind_dims
                                else ss[k] for k in range(idims)]
                        xout = tuple(xout)
                        xind = tuple(xind)

                        xi = list()
                        for kk in dims:
                            if kk in mod_name:
                                # This is the next instrument coordinate
                                k = mod_name.index(kk)
                                if k in imod_dims:
                                    # This is an xarray coordiante
                                    xi.append(inst_coord[kk][cinds[k]])
                                else:
                                    # This is an xarray variable
                                    xi.append(inst_coord[kk][xind])

                        # Cycle the indices
                        if len(cinds) > 0:
                            k = 0
                            cinds[k] += 1

                            while cinds[k] > \
                                inst.data.coords.dims[inst_name[imod_dims[k]]]:
                                k += 1
                                if k < len(cinds):
                                    cinds[k-1] = 0
                                    cinds[k] += 1
                                else:
                                    break
                        icycles += 1

                    # If we have cycled through all the coordinates for this
                    # time, move onto the next time
                    if icycles >= ncycles:
                        get_coords = False

                # Interpolate the desired value
                try:
                    yi = interpolate.interpn(points, values, xi, method=method)
                except ValueError as verr:
                    if str(verr).find("requested xi is out of bounds") > 0:
                        # This is acceptable, pad the interpolated data with
                        # NaN
                        logger.warning("{:} for ".format(verr) +
                              "{:s} data at {:}".format(mdat, xi))
                        yi = [np.nan]
                    else:
                        raise ValueError(verr)

                # Save the output
                attr_name = "{:s}_{:s}".format(model_label, mdat)
                if attr_name not in interp_data.keys():
                    interp_data[attr_name] = np.full(shape=interp_shape,
                                                     fill_value=np.nan)
                interp_data[attr_name][xout] = yi[0]

    # Test and ensure the instrument data doesn't already have the interpolated
    # data.  This should not happen
    if np.any([mdat in inst.data.keys() for mdat in interp_data.keys()]):
        raise ValueError("instrument object already contains model data")

    # Update the instrument object and attach units to the metadata
    for mdat in interp_data.keys():
        attr_name = mdat.split("{:s}_".format(model_label))[-1]
        inst.meta[mdat] = {inst.units_label: model.data_vars[attr_name].units}

        if inst.pandas_format:
            inst[mdat] = pds.Series(interp_data[mdat], index=inst.index)
        else:
            inst.data = inst.data.assign(interp_key=(inst.data.coords.keys(),
                                                     interp_data[mdat]))
            inst.data.rename({"interp_key": mdat}, inplace=True)

    return interp_data.keys()



# -*- coding: utf-8 -*-
"""Supports the Floating Potential Measurement Unit
(FPMU) instrument onboard the International Space
Station (ISS). Downloads data from the NASA
Coordinated Data Analysis Web (CDAWeb).

Properties
----------
platform
    'iss'
name
    'fpmu'
tag
    None Supported
sat_id
    None supported


Warnings
--------
- Currently clean only replaces fill values with Nans.
- Module not written by FPMU team.

"""
from __future__ import print_function
from __future__ import absolute_import
import functools
import numpy as np

import pysat

from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

platform = 'iss'
name = 'fpmu'
tags = {'': ''}
sat_ids = {'': ['']}
_test_dates = {'': {'': pysat.datetime(2017, 10, 1)}}

# support list files routine
# use the default CDAWeb method
fname = 'iss_sp_fpmu_{year:04d}{month:02d}{day:02d}_v01.cdf'
supported_tags = {'': {'': fname}}
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)
# support load routine
# use the default CDAWeb method
load = cdw.load

# support download routine
# use the default CDAWeb method
basic_tag = {'dir': '/pub/data/international_space_station_iss/sp_fpmu',
             'remote_fname': '{year:4d}/'+fname,
             'local_fname': fname}
supported_tags = {'': {'': basic_tag}}
download = functools.partial(cdw.download, supported_tags)

# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


def clean(inst):
    """Return FPMU data cleaned to the specified level.

    Parameters
    ----------
    inst : (pysat.Instrument)
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    """

    inst.data.replace(-999., np.nan, inplace=True)  # Te
    inst.data.replace(-9.9999998e+30, np.nan, inplace=True)  # Ni

    return None



from __future__ import print_function
from __future__ import absolute_import

import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
import warnings

from pysat import logger

def scatterplot(inst, labelx, labely, data_label, datalim, xlim=None,
                ylim=None):
    """Return scatterplot of data_label(s) as functions of labelx,y over a
    season.

    .. deprecated:: 2.2.0
      `scatterplot` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    Parameters
    ----------
    labelx : string
        data product for x-axis
    labely : string
        data product for y-axis
    data_label : string, array-like of strings
        data product(s) to be scatter plotted
    datalim : numyp array
        plot limits for data_label

    Returns
    -------
    Returns a list of scatter plots of data_label as a function
    of labelx and labely over the season delineated by start and
    stop datetime objects.

    """

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    if mpl.is_interactive():
        interactive_mode = True
        # turn interactive plotting off
        plt.ioff()
    else:
        interactive_mode = False

    # create figures for plotting
    figs = []
    axs = []

    # Check for list-like behaviour of data_label
    if type(data_label) is str:
        data_label = [data_label]

    # multiple data to be plotted
    for i in np.arange(len(data_label)):
        figs.append(plt.figure())
        ax1 = figs[i].add_subplot(211, projection='3d')
        ax2 = figs[i].add_subplot(212)
        axs.append((ax1, ax2))
        plt.suptitle(data_label[i])
        if xlim is not None:
            ax1.set_xlim(xlim)
            ax2.set_xlim(xlim)
        if ylim is not None:
            ax1.set_ylim(ylim)
            ax2.set_ylim(ylim)

    # norm method so that data may be scaled to colors appropriately
    norm = mpl.colors.Normalize(vmin=datalim[0], vmax=datalim[1])
    p = [i for i in np.arange(len(figs))]
    q = [i for i in np.arange(len(figs))]
    for i, inst in enumerate(inst):
        for j, (fig, ax) in enumerate(zip(figs, axs)):
            if not inst.empty:
                check1 = len(inst.data[labelx]) > 0
                check2 = len(inst.data[labely]) > 0
                check3 = len(inst.data[data_label[j]]) > 0
                if (check1 & check2 & check3):
                    p[j] = ax[0].scatter(inst.data[labelx], inst.data[labely],
                                         inst.data[data_label[j]], zdir='z',
                                         c=inst.data[data_label[j]], norm=norm,
                                         linewidth=0, edgecolors=None)
                    q[j] = ax[1].scatter(inst.data[labelx], inst.data[labely],
                                         c=inst.data[data_label[j]],
                                         norm=norm, alpha=0.5, edgecolor=None)

    for j, (fig, ax) in enumerate(zip(figs, axs)):
        try:
            plt.colorbar(p[j], ax=ax[0], label='Amplitude (m/s)')
        except:
            logger.info('Tried colorbar but failed, thus no colorbar.')
        ax[0].elev = 30.

    if interactive_mode:
        # turn interactive plotting back on
        plt.ion()

    return figs



"""
tests the pysat averaging code
"""
from nose.tools import raises
import numpy as np
import pandas as pds
import warnings
import pysat
from pysat.ssnl import avg


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean')
        self.bounds1 = (pysat.datetime(2008, 1, 1), pysat.datetime(2008, 1, 3))
        self.bounds2 = (pysat.datetime(2009, 1, 1), pysat.datetime(2009, 1, 2))

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.bounds1, self.bounds2

    def test_basic_seasonal_median2D(self):
        """ Test the basic seasonal 2D median"""
        self.testInst.bounds = self.bounds1
        results = avg.median2D(self.testInst, [0., 360., 24], 'longitude',
                               [0., 24, 24], 'mlt',
                               ['dummy1', 'dummy2', 'dummy3'])
        dummy_val = results['dummy1']['median']
        dummy_dev = results['dummy1']['avg_abs_dev']

        dummy2_val = results['dummy2']['median']
        dummy2_dev = results['dummy2']['avg_abs_dev']

        dummy3_val = results['dummy3']['median']
        dummy3_dev = results['dummy3']['avg_abs_dev']

        dummy_x = results['dummy1']['bin_x']
        dummy_y = results['dummy1']['bin_y']

        # iterate over all y rows
        # value should be equal to integer value of mlt
        # no variation in the median, all values should be the same
        check = []
        for i, y in enumerate(dummy_y[:-1]):
            assert np.all(dummy_val[i, :] == y.astype(int))
            assert np.all(dummy_dev[i, :] == 0)

        for i, x in enumerate(dummy_x[:-1]):
            assert np.all(dummy2_val[:, i] == x/15.0)
            assert np.all(dummy2_dev[:, i] == 0)

        for i, x in enumerate(dummy_x[:-1]):
            check.append(np.all(dummy3_val[:, i] == x/15.0 * 1000.0
                                + dummy_y[:-1]))
            check.append(np.all(dummy3_dev[:, i] == 0))

        # holds here because there are 32 days, no data is discarded,
        # each day holds same amount of data
        assert(self.testInst.data['dummy1'].size*3 ==
               sum([sum(i) for i in results['dummy1']['count']]))

        assert np.all(check)

    def test_basic_daily_mean(self):
        """ Test basic daily mean"""
        self.testInst.bounds = self.bounds1
        ans = avg.mean_by_day(self.testInst, 'dummy4')
        assert np.all(ans == 86399/2.0)

    def test_basic_orbit_mean(self):
        """Test basic orbital mean"""
        orbit_info = {'kind': 'local time', 'index': 'mlt'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=orbit_info)
        self.testInst.bounds = self.bounds2
        ans = avg.mean_by_orbit(self.testInst, 'mlt')
        # note last orbit is incomplete thus not expected to satisfy relation
        assert np.allclose(ans[:-1], np.ones(len(ans)-1)*12.0, 1.0E-2)

    def test_basic_file_mean(self):
        """Test basic file mean"""
        index = pds.date_range(*self.bounds1)
        names = [date.strftime('%Y-%m-%d')+'.nofile' for date in index]
        self.testInst.bounds = (names[0], names[-1])
        ans = avg.mean_by_file(self.testInst, 'dummy4')
        assert np.all(ans == 86399/2.0)


class TestDeprecation():

    def setup(self):
        """Runs before every method to create a clean testing setup"""
        warnings.simplefilter("always")

    def teardown(self):
        """Runs after every method to clean up previous testing"""

    def test_median1D_deprecation_warning(self):
        """Test generation of deprecation warning for median1D"""

        with warnings.catch_warnings(record=True) as war:
            try:
                avg.median1D(None, [0., 360., 24],
                             'longitude', ['dummy1'])
            except ValueError:
                # Setting inst to None should produce a ValueError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_median2D_deprecation_warning(self):
        """Test generation of deprecation warning for median1D"""

        with warnings.catch_warnings(record=True) as war:
            try:
                avg.median2D(None, [0., 360., 24], 'longitude',
                             [0., 24., 24], 'mlt', ['dummy1'])
            except ValueError:
                # Setting inst to None should produce a ValueError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_mean_by_day_deprecation_warning(self):
        """Test generation of deprecation warning for mean_by_day"""

        with warnings.catch_warnings(record=True) as war:
            try:
                avg.mean_by_day(None, 'dummy1')
            except TypeError:
                # Setting inst to None should produce a TypeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_mean_by_orbit_deprecation_warning(self):
        """Test generation of deprecation warning for mean_by_orbit"""

        with warnings.catch_warnings(record=True) as war:
            try:
                avg.mean_by_orbit(None, 'dummy1')
            except AttributeError:
                # Setting inst to None should produce a AttributeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_mean_by_file_deprecation_warning(self):
        """Test generation of deprecation warning for mean_by_file"""

        with warnings.catch_warnings(record=True) as war:
            try:
                avg.mean_by_file(None, 'dummy1')
            except TypeError:
                # Setting inst to None should produce a TypeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning


class TestFrameProfileAverages():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat', name='testing2D',
                                         clean_level='clean')
        self.testInst.bounds = (pysat.datetime(2008, 1, 1),
                                pysat.datetime(2008, 1, 3))
        self.dname = 'alt_profiles'
        self.test_vals = np.arange(50) * 1.2
        self.test_fracs = np.arange(50) / 50.0

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.dname, self.test_vals, self.test_fracs

    def test_basic_seasonal_2Dmedian(self):
        """ Test the basic seasonal 2D median"""

        results = avg.median2D(self.testInst, [0., 360., 24], 'longitude',
                               [0., 24, 24], 'mlt', [self.dname])

        # iterate over all
        # no variation in the median, all values should be the same
        for i, row in enumerate(results[self.dname]['median']):
            for j, item in enumerate(row):
                assert np.all(item['density'] == self.test_vals)
                assert np.all(item['fraction'] == self.test_fracs)

        for i, row in enumerate(results[self.dname]['avg_abs_dev']):
            for j, item in enumerate(row):
                assert np.all(item['density'] == 0)
                assert np.all(item['fraction'] == 0)

    def test_basic_seasonal_1Dmedian(self):
        """ Test the basic seasonal 1D median"""

        results = avg.median1D(self.testInst, [0., 24, 24], 'mlt',
                               [self.dname])

        # iterate over all
        # no variation in the median, all values should be the same
        for i, row in enumerate(results[self.dname]['median']):
            assert np.all(row['density'] == self.test_vals)
            assert np.all(row['fraction'] == self.test_fracs)

        for i, row in enumerate(results[self.dname]['avg_abs_dev']):
            assert np.all(row['density'] == 0)
            assert np.all(row['fraction'] == 0)


class TestSeriesProfileAverages():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat', name='testing2D',
                                         clean_level='clean')
        self.testInst.bounds = (pysat.datetime(2008, 1, 1),
                                pysat.datetime(2008, 2, 1))
        self.dname = 'series_profiles'

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.dname

    def test_basic_seasonal_median2D(self):
        """ Test basic seasonal 2D median"""
        results = avg.median2D(self.testInst, [0., 360., 24], 'longitude',
                               [0., 24, 24], 'mlt', [self.dname])

        # iterate over all
        # no variation in the median, all values should be the same
        test_vals = np.arange(50) * 1.2
        for i, row in enumerate(results[self.dname]['median']):
            for j, item in enumerate(row):
                assert np.all(item == test_vals)

        for i, row in enumerate(results[self.dname]['avg_abs_dev']):
            for j, item in enumerate(row):
                assert np.all(item == 0)

    def test_basic_seasonal_median1D(self):
        """ Test basic seasonal 1D median"""
        results = avg.median1D(self.testInst, [0., 24, 24], 'mlt',
                               [self.dname])

        # iterate over all
        # no variation in the median, all values should be the same
        test_vals = np.arange(50) * 1.2
        for i, row in enumerate(results[self.dname]['median']):
            assert np.all(row == test_vals)

        for i, row in enumerate(results[self.dname]['avg_abs_dev']):
            assert np.all(row == 0)


class TestConstellation:
    def setup(self):
        insts = []
        for i in range(5):
            insts.append(pysat.Instrument('pysat', 'testing',
                                          clean_level='clean'))
        self.testC = pysat.Constellation(instruments=insts)
        self.testI = pysat.Instrument('pysat', 'testing', clean_level='clean')
        self.bounds = (pysat.datetime(2008, 1, 1), pysat.datetime(2008, 1, 3))

    def teardown(self):
        del self.testC, self.testI, self.bounds

    def test_constellation_median2D(self):
        """ Test constellation implementation of 2D median"""
        for i in self.testC.instruments:
            i.bounds = self.bounds
        self.testI.bounds = self.bounds
        resultsC = avg.median2D(self.testC, [0., 360., 24], 'longitude',
                                [0., 24, 24], 'mlt',
                                ['dummy1', 'dummy2', 'dummy3'])
        resultsI = avg.median2D(self.testI, [0., 360., 24], 'longitude',
                                [0., 24, 24], 'mlt',
                                ['dummy1', 'dummy2', 'dummy3'])
        medC1 = resultsC['dummy1']['median']
        medI1 = resultsI['dummy1']['median']
        medC2 = resultsC['dummy2']['median']
        medI2 = resultsI['dummy2']['median']
        medC3 = resultsC['dummy3']['median']
        medI3 = resultsI['dummy3']['median']

        assert np.array_equal(medC1, medI1)
        assert np.array_equal(medC2, medI2)
        assert np.array_equal(medC3, medI3)

    def test_constellation_median1D(self):
        """ Test constellation implementation of 1D median"""
        for i in self.testC.instruments:
            i.bounds = self.bounds
        self.testI.bounds = self.bounds
        resultsC = avg.median1D(self.testC, [0., 24, 24], 'mlt',
                                ['dummy1', 'dummy2', 'dummy3'])
        resultsI = avg.median1D(self.testI, [0., 24, 24], 'mlt',
                                ['dummy1', 'dummy2', 'dummy3'])
        medC1 = resultsC['dummy1']['median']
        medI1 = resultsI['dummy1']['median']
        medC2 = resultsC['dummy2']['median']
        medI2 = resultsI['dummy2']['median']
        medC3 = resultsC['dummy3']['median']
        medI3 = resultsI['dummy3']['median']

        assert np.array_equal(medC1, medI1)
        assert np.array_equal(medC2, medI2)
        assert np.array_equal(medC3, medI3)


class TestHeterogenousConstellation:
    def setup(self):
        insts = []
        for i in range(2):
            r_date = pysat.datetime(2009, 1, i+1)
            insts.append(pysat.Instrument('pysat', 'testing',
                                          clean_level='clean',
                                          root_date=r_date))
        self.testC = pysat.Constellation(instruments=insts)
        self.bounds = (pysat.datetime(2008, 1, 1), pysat.datetime(2008, 1, 3))

    def teardown(self):
        del self.testC, self.bounds

    def test_heterogenous_constellation_median2D(self):
        """ Test the seasonal 2D median of a heterogeneous constellation """
        for inst in self.testC:
            inst.bounds = self.bounds
        results = avg.median2D(self.testC, [0., 360., 24], 'longitude',
                               [0., 24, 24], 'mlt',
                               ['dummy1', 'dummy2', 'dummy3'])
        dummy_val = results['dummy1']['median']
        dummy_dev = results['dummy1']['avg_abs_dev']

        dummy2_val = results['dummy2']['median']
        dummy2_dev = results['dummy2']['avg_abs_dev']

        dummy3_val = results['dummy3']['median']
        dummy3_dev = results['dummy3']['avg_abs_dev']

        dummy_x = results['dummy1']['bin_x']
        dummy_y = results['dummy1']['bin_y']

        # iterate over all y rows
        # value should be equal to integer value of mlt
        # no variation in the median, all values should be the same
        check = []
        for i, y in enumerate(dummy_y[:-1]):
            check.append(np.all(dummy_val[i, :] == y.astype(int)))
            check.append(np.all(dummy_dev[i, :] == 0))

        for i, x in enumerate(dummy_x[:-1]):
            check.append(np.all(dummy2_val[:, i] == x/15.0))
            check.append(np.all(dummy2_dev[:, i] == 0))

        for i, x in enumerate(dummy_x[:-1]):
            check.append(np.all(dummy3_val[:, i] == x/15.0 * 1000.0
                                + dummy_y[:-1]))
            check.append(np.all(dummy3_dev[:, i] == 0))

        assert np.all(check)

    def test_heterogenous_constellation_median1D(self):
        """ Test the seasonal 1D median of a heterogeneous constellation """
        for inst in self.testC:
            inst.bounds = self.bounds
        results = avg.median1D(self.testC, [0., 24, 24], 'mlt', ['dummy1'])

        # Extract the results
        dummy_val = results['dummy1']['median']
        dummy_dev = results['dummy1']['avg_abs_dev']

        # iterate over all x rows
        # value should be equal to integer value of mlt
        # no variation in the median, all values should be the same
        check = []
        for i, x in enumerate(results['dummy1']['bin_x'][:-1]):
            check.append(np.all(dummy_val[i] == x.astype(int)))
            check.append(np.all(dummy_dev[i] == 0))

        assert np.all(check)


class Test2DConstellation:
    def setup(self):
        insts = []
        insts.append(pysat.Instrument(platform='pysat', name='testing2D',
                                      clean_level='clean'))
        self.testC = pysat.Constellation(insts)
        self.bounds = (pysat.datetime(2008, 1, 1), pysat.datetime(2008, 1, 3))

    def teardown(self):
        del self.testC, self.bounds

    def test_2D_median(self):
        """ Test a 2D median calculation with a constellation"""
        for i in self.testC.instruments:
            i.bounds = self.bounds

        results = avg.median2D(self.testC, [0., 360., 24], 'longitude',
                               [0., 24, 24], 'slt', ['uts'])
        dummy_val = results['uts']['median']
        dummy_dev = results['uts']['avg_abs_dev']

        dummy_y = results['uts']['bin_y']

        # iterate over all y rows
        # value should be equal to integer value of mlt
        # no variation in the median, all values should be the same
        check = []
        for i, y in enumerate(dummy_y[:-1]):
            check.append(np.all(dummy_val[i, :] == y.astype(int)))
            check.append(np.all(dummy_dev[i, :] == 0))

    def test_1D_median(self):
        """ Test a 1D median calculation with a constellation"""
        for i in self.testC.instruments:
            i.bounds = self.bounds

        results = avg.median1D(self.testC, [0., 24, 24], 'slt', ['uts'])
        dummy_val = results['uts']['median']
        dummy_dev = results['uts']['avg_abs_dev']

        dummy_x = results['uts']['bin_x']

        # iterate over all x rows
        # value should be equal to integer value of slt
        # no variation in the median, all values should be the same
        check = []
        for i, x in enumerate(dummy_x[:-1]):
            check.append(np.all(dummy_val[i] == x.astype(int)))
            check.append(np.all(dummy_dev[i] == 0))


class TestSeasonalAverageUnevenBins:
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean')
        self.testInst.bounds = (pysat.datetime(2008, 1, 1),
                                pysat.datetime(2008, 1, 3))

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst

    def test_seasonal_average_uneven_bins(self):
        """ Test seasonal 2D median with uneven bins"""
        results = avg.median2D(self.testInst, np.linspace(0., 360., 25),
                               'longitude', np.linspace(0., 24., 25), 'mlt',
                               ['dummy1', 'dummy2', 'dummy3'], auto_bin=False)
        dummy_val = results['dummy1']['median']
        dummy_dev = results['dummy1']['avg_abs_dev']

        dummy2_val = results['dummy2']['median']
        dummy2_dev = results['dummy2']['avg_abs_dev']

        dummy3_val = results['dummy3']['median']
        dummy3_dev = results['dummy3']['avg_abs_dev']

        dummy_x = results['dummy1']['bin_x']
        dummy_y = results['dummy1']['bin_y']

        # iterate over all y rows
        # value should be equal to integer value of mlt
        # no variation in the median, all values should be the same
        check = []
        for i, y in enumerate(dummy_y[:-1]):
            assert np.all(dummy_val[i, :] == y.astype(int))
            assert np.all(dummy_dev[i, :] == 0)

        for i, x in enumerate(dummy_x[:-1]):
            assert np.all(dummy2_val[:, i] == x/15.0)
            assert np.all(dummy2_dev[:, i] == 0)

        for i, x in enumerate(dummy_x[:-1]):
            check.append(np.all(dummy3_val[:, i] == x/15.0 * 1000.0
                                + dummy_y[:-1]))
            check.append(np.all(dummy3_dev[:, i] == 0))

        # holds here because there are 32 days, no data is discarded,
        # each day holds same amount of data
        assert(self.testInst.data['dummy1'].size*3 ==
               sum([sum(i) for i in results['dummy1']['count']]))

        assert np.all(check)

    @raises(ValueError)
    def test_nonmonotonic_bins(self):
        """Test 2D median failure when provided with a non-monotonic bins
        """
        avg.median2D(self.testInst, np.array([0., 300., 100.]), 'longitude',
                     np.array([0., 24., 13.]), 'mlt',
                     ['dummy1', 'dummy2', 'dummy3'], auto_bin=False)

    @raises(TypeError)
    def test_bin_data_depth(self):
        """Test failure when an array-like of length 1 is given to median2D
        """
        avg.median2D(self.testInst, 1, 'longitude', 24, 'mlt',
                     ['dummy1', 'dummy2', 'dummy3'], auto_bin=False)

    @raises(TypeError)
    def test_bin_data_type(self):
        """Test failure when a non array-like is given to median2D
        """
        avg.median2D(self.testInst, ['1', 'a', '23', '10'], 'longitude',
                     ['0', 'd', '24', 'c'], 'mlt',
                     ['dummy1', 'dummy2', 'dummy3'], auto_bin=False)


class TestInstMed1D():
    def setup(self):
        """Runs before every method to create a clean testing setup"""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         update_files=True)
        self.testInst.bounds = (pysat.datetime(2008, 1, 1),
                                pysat.datetime(2008, 1, 31))
        self.test_bins = [0, 24, 24]
        self.test_label = 'slt'
        self.test_data = ['dummy1', 'dummy2']
        self.out_keys = ['count', 'avg_abs_dev', 'median', 'bin_x']
        self.out_data = {'dummy1':
                         {'count': [111780., 111320., 111780., 111320.,
                                    111780., 111320., 111780., 111320.,
                                    111780., 111320., 111780., 111320.,
                                    111780., 111320., 111918., 111562.,
                                    112023., 111562., 112023., 111412.,
                                    111780., 111320., 111780., 111320.],
                          'avg_abs_dev': np.zeros(shape=24),
                          'median': np.linspace(0.0, 23.0, 24)},
                         'dummy2':
                         {'count': [111780., 111320., 111780., 111320.,
                                    111780., 111320., 111780., 111320.,
                                    111780., 111320., 111780., 111320.,
                                    111780., 111320., 111918., 111562.,
                                    112023., 111562., 112023., 111412.,
                                    111780., 111320., 111780., 111320.],
                          'avg_abs_dev': np.zeros(shape=24) + 6.0,
                          'median': [11., 12., 11., 11., 12., 11., 12., 11.,
                                     12., 12., 11., 12., 11., 12., 11., 11.,
                                     12., 11., 12., 11., 11., 11., 11., 12.]}}

    def teardown(self):
        """Runs after every method to clean up previous testing"""
        del self.testInst, self.test_bins, self.test_label, self.test_data
        del self.out_keys, self.out_data

    def test_median1D_default(self):
        """Test success of median1D with default options"""

        med_dict = avg.median1D(self.testInst, self.test_bins, self.test_label,
                                self.test_data)

        # Test output type
        assert isinstance(med_dict, dict)
        assert len(med_dict.keys()) == len(self.test_data)

        # Test output keys
        for kk in med_dict.keys():
            assert kk in self.test_data
            assert np.all([jj in self.out_keys
                           for jj in med_dict[kk].keys()])

            # Test output values
            for jj in self.out_keys[:-1]:
                assert len(med_dict[kk][jj]) == self.test_bins[-1]
                assert np.all(med_dict[kk][jj] == self.out_data[kk][jj])

            jj = self.out_keys[-1]
            assert len(med_dict[kk][jj]) == self.test_bins[-1]+1
            assert np.all(med_dict[kk][jj] == np.linspace(self.test_bins[0],
                                                          self.test_bins[1],
                                                          self.test_bins[2]+1))
        del med_dict, kk, jj

    @raises(KeyError)
    def test_median1D_bad_data(self):
        """Test failure of median1D with string data instead of list"""

        avg.median1D(self.testInst, self.test_bins, self.test_label,
                     self.test_data[0])

    @raises(KeyError)
    def test_median1D_bad_label(self):
        """Test failure of median1D with unknown label"""

        avg.median1D(self.testInst, self.test_bins, "bad_label",
                     self.test_data)

    @raises(ValueError)
    def test_nonmonotonic_bins(self):
        """Test median1D failure when provided with a non-monotonic bins
        """
        avg.median1D(self.testInst, [0, 13, 5], self.test_label,
                     self.test_data, auto_bin=False)

    @raises(TypeError)
    def test_bin_data_depth(self):
        """Test failure when array-like of length 1 is given to median1D
        """
        avg.median1D(self.testInst, 24, self.test_label, self.test_data,
                     auto_bin=False)

    @raises(TypeError)
    def test_bin_data_type(self):
        """Test failure when median 1D is given non array-like bins
        """
        pysat.ssnl.avg.median2D(self.testInst, ['0', 'd', '24', 'c'],
                                self.test_label, self.test_data,
                                auto_bin=False)



from __future__ import print_function
from __future__ import absolute_import

import numpy as np
import warnings

import xarray as xr

import pysat


def set_data_dir(path=None, store=True):
    """
    Set the top level directory pysat uses to look for data and reload.

    Parameters
    ----------
    path : string
        valid path to directory pysat uses to look for data
    store : bool
        if True, store data directory for future runs
    """

    import os
    import sys
    import pysat
    if sys.version_info[0] >= 3:
        from importlib import reload as re_load
    else:
        re_load = reload

    # account for a user prefix in the path, such as ~
    path = os.path.expanduser(path)
    # account for the presence of $HOME or similar
    path = os.path.expandvars(path)

    if os.path.isdir(path):
        if store:
            with open(os.path.join(os.path.expanduser('~'), '.pysat',
                                   'data_path.txt'), 'w') as f:
                f.write(path)
        pysat.data_dir = path
        pysat._files = re_load(pysat._files)
        pysat._instrument = re_load(pysat._instrument)
    else:
        raise ValueError('Path %s does not lead to a valid directory.' % path)


def computational_form(data):
    """Deprecated function.  Moved to pysat.ssnl.computational_form"""

    import warnings

    warnings.warn(' '.join(["utils.computational_form is deprecated, use",
                            "pysat.ssnl.computational_form instead"]),
                  DeprecationWarning, stacklevel=2)
    dslice = pysat.ssnl.computational_form(data)

    return dslice


def scale_units(out_unit, in_unit):
    """ Determine the scaling factor between two units

    Parameters
    -------------
    out_unit : str
        Desired unit after scaling
    in_unit : str
        Unit to be scaled

    Returns
    -----------
    unit_scale : float
        Scaling factor that will convert from in_units to out_units

    Notes
    -------
    Accepted units include degrees ('deg', 'degree', 'degrees'),
    radians ('rad', 'radian', 'radians'),
    hours ('h', 'hr', 'hrs', 'hour', 'hours'), and lengths ('m', 'km', 'cm').
    Can convert between degrees, radians, and hours or different lengths.

    Example
    -----------
    ::
    import numpy as np
    two_pi = 2.0 * np.pi
    scale = scale_units("deg", "RAD")
    two_pi *= scale
    two_pi # will show 360.0


    """

    if out_unit == in_unit:
        return 1.0

    accepted_units = {'deg': ['deg', 'degree', 'degrees'],
                      'rad': ['rad', 'radian', 'radians'],
                      'h': ['h', 'hr', 'hrs', 'hours'],
                      'm': ['m', 'km', 'cm'],
                      'm/s': ['m/s', 'cm/s', 'km/s', 'm s$^{-1}$',
                              'cm s$^{-1}$', 'km s$^{-1}$', 'm s-1', 'cm s-1',
                              'km s-1']}
    replace_str = {'/s': [' s$^{-1}$', ' s-1']}

    scales = {'deg': 180.0, 'rad': np.pi, 'h': 12.0,
              'm': 1.0, 'km': 0.001, 'cm': 100.0,
              'm/s': 1.0, 'cm/s': 100.0, 'km/s': 0.001}

    # Test input and determine transformation type
    out_key = out_unit.lower()
    in_key = in_unit.lower()
    for kk in accepted_units.keys():
        if out_key in accepted_units.keys() and in_key in accepted_units.keys():
            break

        if(out_key not in accepted_units.keys() and
           out_unit.lower() in accepted_units[kk]):
            out_key = kk
        if(in_key not in accepted_units.keys() and
           in_unit.lower() in accepted_units[kk]):
            in_key = kk

    if(out_key not in accepted_units.keys() and
       in_key not in accepted_units.keys()):
        raise ValueError(''.join(['Cannot scale {:s} and '.format(in_unit),
                                  '{:s}, unknown units'.format(out_unit)]))

    if out_key not in accepted_units.keys():
        raise ValueError('Unknown output unit {:}'.format(out_unit))

    if in_key not in accepted_units.keys():
        raise ValueError('Unknown input unit {:}'.format(in_unit))

    if out_key == 'm' or out_key == 'm/s' or in_key == 'm' or in_key == 'm/s':
        if in_key != out_key:
            raise ValueError('Cannot scale {:s} and {:s}'.format(out_unit,
                                                                 in_unit))
        # Recast units as keys for the scales dictionary and ensure that
        # the format is consistent
        rkey = ''
        for rr in replace_str.keys():
            if out_key.find(rr):
                rkey = rr

        out_key = out_unit.lower()
        in_key = in_unit.lower()

        if rkey in replace_str.keys():
            for rval in replace_str[rkey]:
                out_key = out_key.replace(rval, rkey)
                in_key = in_key.replace(rval, rkey)

    unit_scale = scales[out_key] / scales[in_key]

    return unit_scale


def load_netcdf4(fnames=None, strict_meta=False, file_format=None,
                 epoch_name='Epoch', units_label='units',
                 name_label='long_name', notes_label='notes',
                 desc_label='desc', plot_label='label', axis_label='axis',
                 scale_label='scale', min_label='value_min',
                 max_label='value_max', fill_label='fill',
                 pandas_format=True):
    # unix_time=False, **kwargs):
    """Load netCDF-3/4 file produced by pysat.

    Parameters
    ----------
    fnames : string or array_like of strings (None)
        filenames to load
    strict_meta : boolean (False)
        check if metadata across fnames is the same
    file_format : string (None)
        file_format keyword passed to netCDF4 routine
        NETCDF3_CLASSIC, NETCDF3_64BIT, NETCDF4_CLASSIC, and NETCDF4
    epoch_name : string ('Epoch')
    units_label : string ('units')
        keyword for unit information
    name_label : string ('long_name')
        keyword for informative name label
    notes_label : string ('notes')
        keyword for file notes
    desc_label : string ('desc')
        keyword for data descriptions
    plot_label : string ('label')
        keyword for name to use on plot labels
    axis_label : string ('axis')
        keyword for axis labels
    scale_label : string ('scale')
        keyword for plot scaling
    min_label : string ('value_min')
        keyword for minimum in allowable value range
    max_label : string ('value_max')
        keyword for maximum in allowable value range
    fill_label : string ('fill')
        keyword for fill values

    Returns
    --------
    out : pandas.core.frame.DataFrame
        DataFrame output
    mdata : pysat._meta.Meta
        Meta data
    """

    import copy
    import netCDF4
    import pandas as pds
    import pysat
    try:
        basestring
    except NameError:
        basestring = str

    if fnames is None:
        raise ValueError("Must supply a filename/list of filenames")
    if isinstance(fnames, basestring):
        fnames = [fnames]

    if file_format is None:
        file_format = 'NETCDF4'
    else:
        file_format = file_format.upper()

    saved_mdata = None
    running_idx = 0
    running_store = []
    two_d_keys = []
    two_d_dims = []
    three_d_keys = []
    three_d_dims = []
    mdata = pysat.Meta(units_label=units_label,
                       name_label=name_label,
                       notes_label=notes_label,
                       desc_label=desc_label,
                       plot_label=plot_label,
                       axis_label=axis_label,
                       scale_label=scale_label,
                       min_label=min_label,
                       max_label=max_label,
                       fill_label=fill_label)

    if pandas_format:
        for fname in fnames:
            with netCDF4.Dataset(fname, mode='r', format=file_format) as data:
                # build up dictionary with all global ncattrs
                # and add those attributes to a pysat meta object
                ncattrsList = data.ncattrs()
                for d in ncattrsList:
                    if hasattr(mdata, d):
                        mdata.__setattr__(d+'_', data.getncattr(d))
                    else:
                        mdata.__setattr__(d, data.getncattr(d))

                loadedVars = {}
                for key in data.variables.keys():
                    # load up metadata.  From here group unique
                    # dimensions and act accordingly, 1D, 2D, 3D
                    if len(data.variables[key].dimensions) == 1:
                        if pandas_format:
                            # load 1D data variable
                            # assuming basic time dimension
                            loadedVars[key] = data.variables[key][:]
                        # load up metadata
                        meta_dict = {}
                        for nc_key in data.variables[key].ncattrs():
                            meta_dict[nc_key] = \
                                    data.variables[key].getncattr(nc_key)
                        mdata[key] = meta_dict
                    if len(data.variables[key].dimensions) == 2:
                        # part of dataframe within dataframe
                        two_d_keys.append(key)
                        two_d_dims.append(data.variables[key].dimensions)

                    if len(data.variables[key].dimensions) == 3:
                        warnings.warn(' '.join(["Support for 3D data in pandas",
                                                "will be removed in pysat 3.0",
                                                "Please use xarray for",
                                                "multi-dimension data."]),
                                      DeprecationWarning, stacklevel=2)
                        # part of full/dedicated dataframe within dataframe
                        three_d_keys.append(key)
                        three_d_dims.append(data.variables[key].dimensions)

                # we now have a list of keys that need to go into a dataframe,
                # could be more than one, collect unique dimensions for 2D keys
                for dim in set(two_d_dims):
                    # first or second dimension could be epoch
                    # Use other dimension name as variable name
                    if dim[0] == epoch_name:
                        obj_key_name = dim[1]
                    elif dim[1] == epoch_name:
                        obj_key_name = dim[0]
                    else:
                        raise KeyError('Epoch not found!')
                    # collect variable names associated with dimension
                    idx_bool = [dim == i for i in two_d_dims]
                    idx, = np.where(np.array(idx_bool))
                    obj_var_keys = []
                    clean_var_keys = []
                    for i in idx:
                        obj_var_keys.append(two_d_keys[i])
                        clean_var_keys.append(
                                two_d_keys[i].split(obj_key_name + '_')[-1])

                    # figure out how to index this data, it could provide its
                    # own index - or we may have to create simple integer based
                    # DataFrame access. If the dimension is stored as its own
                    # variable then use that info for index
                    if obj_key_name in obj_var_keys:
                        # string used to indentify dimension also in
                        # data.variables will be used as an index
                        index_key_name = obj_key_name
                        # if the object index uses UNIX time, process into
                        # datetime index
                        if data.variables[obj_key_name].getncattr(name_label) == \
                                epoch_name:
                            # name to be used in DataFrame index
                            index_name = epoch_name
                            time_index_flag = True
                        else:
                            time_index_flag = False
                            # label to be used in DataFrame index
                            index_name = \
                                data.variables[obj_key_name].getncattr(name_label)
                    else:
                        # dimension is not itself a variable
                        index_key_name = None

                    # iterate over the variables and grab metadata
                    dim_meta_data = pysat.Meta(units_label=units_label,
                                               name_label=name_label,
                                               notes_label=notes_label,
                                               desc_label=desc_label,
                                               plot_label=plot_label,
                                               axis_label=axis_label,
                                               scale_label=scale_label,
                                               min_label=min_label,
                                               max_label=max_label,
                                               fill_label=fill_label)

                    for key, clean_key in zip(obj_var_keys, clean_var_keys):
                        # store attributes in metadata, exept for dim name
                        meta_dict = {}
                        for nc_key in data.variables[key].ncattrs():
                            meta_dict[nc_key] = \
                                data.variables[key].getncattr(nc_key)
                        dim_meta_data[clean_key] = meta_dict

                    dim_meta_dict = {'meta': dim_meta_data}
                    if index_key_name is not None:
                        # add top level meta
                        for nc_key in data.variables[obj_key_name].ncattrs():
                            dim_meta_dict[nc_key] = \
                                data.variables[obj_key_name].getncattr(nc_key)
                        mdata[obj_key_name] = dim_meta_dict

                    # iterate over all variables with this dimension
                    # data storage, whole shebang
                    loop_dict = {}
                    # list holds a series of slices, parsed from dict above
                    loop_list = []
                    for key, clean_key in zip(obj_var_keys, clean_var_keys):
                        # data
                        loop_dict[clean_key] = \
                            data.variables[key][:, :].flatten(order='C')
                    # number of values in time
                    loop_lim = data.variables[obj_var_keys[0]].shape[0]
                    # number of values per time
                    step_size = len(data.variables[obj_var_keys[0]][0, :])
                    # check if there is an index we should use
                    if not (index_key_name is None):
                        # an index was found
                        time_var = loop_dict.pop(index_key_name)
                        if time_index_flag:
                            # create datetime index from data
                            time_var = pds.to_datetime(1E6 * time_var)
                        new_index = time_var
                        new_index_name = index_name
                    else:
                        # using integer indexing
                        new_index = np.arange(loop_lim*step_size,
                                              dtype=int) % step_size
                        new_index_name = 'index'
                    # load all data into frame
                    if len(loop_dict.keys()) > 1:
                        loop_frame = pds.DataFrame(loop_dict,
                                                   columns=clean_var_keys)
                        if obj_key_name in loop_frame:
                            del loop_frame[obj_key_name]
                        # break massive frame into bunch of smaller frames
                        for i in np.arange(loop_lim, dtype=int):
                            loop_list.append(loop_frame.iloc[step_size*i:step_size*(i+1), :])
                            loop_list[-1].index = new_index[step_size*i:step_size*(i+1)]
                            loop_list[-1].index.name = new_index_name
                    else:
                        loop_frame = pds.Series(loop_dict[clean_var_keys[0]],
                                                name=obj_var_keys[0])
                        # break massive series into bunch of smaller series
                        for i in np.arange(loop_lim, dtype=int):
                            loop_list.append(loop_frame.iloc[step_size*i:step_size*(i+1)])
                            loop_list[-1].index = new_index[step_size*i:step_size*(i+1)]
                            loop_list[-1].index.name = new_index_name

                    # add 2D object data, all based on a unique dimension within
                    # netCDF, to loaded data dictionary
                    loadedVars[obj_key_name] = loop_list
                    del loop_list

                # we now have a list of keys that need to go into a dataframe,
                # could be more than one, collect unique dimensions for 2D keys
                for dim in set(three_d_dims):
                    # collect variable names associated with dimension
                    idx_bool = [dim == i for i in three_d_dims]
                    idx, = np.where(np.array(idx_bool))
                    obj_var_keys = []
                    for i in idx:
                        obj_var_keys.append(three_d_keys[i])

                    for obj_key_name in obj_var_keys:
                        # store attributes in metadata
                        meta_dict = {}
                        for nc_key in data.variables[obj_key_name].ncattrs():
                            meta_dict[nc_key] = \
                                data.variables[obj_key_name].getncattr(nc_key)
                        mdata[obj_key_name] = meta_dict

                        # iterate over all variables with this dimension and store
                        # data
                        loop_dict = {}
                        # list holds a series of slices, parsed from dict above
                        loop_list = []
                        loop_dict[obj_key_name] = \
                            data.variables[obj_key_name][:, :, :]
                        # number of values in time
                        loop_lim = data.variables[obj_key_name].shape[0]
                        # number of values per time
                        step_size_x = len(data.variables[obj_key_name][0, :, 0])
                        step_size_y = len(data.variables[obj_key_name][0, 0, :])
                        step_size = step_size_x
                        loop_dict[obj_key_name] = \
                            loop_dict[obj_key_name].reshape((loop_lim*step_size_x,
                                                             step_size_y))
                        # check if there is an index we should use
                        if not (index_key_name is None):
                            # an index was found
                            time_var = loop_dict.pop(index_key_name)
                            if time_index_flag:
                                # create datetime index from data
                                time_var = pds.to_datetime(1E6 * time_var)
                            new_index = time_var
                            new_index_name = index_name
                        else:
                            # using integer indexing
                            new_index = np.arange(loop_lim*step_size,
                                                  dtype=int) % step_size
                            new_index_name = 'index'
                        # load all data into frame
                        loop_frame = pds.DataFrame(loop_dict[obj_key_name])
                        # del loop_frame['dimension_1']
                        # break massive frame into bunch of smaller frames
                        for i in np.arange(loop_lim, dtype=int):
                            loop_list.append(loop_frame.iloc[step_size*i:step_size*(i+1), :])
                            loop_list[-1].index = new_index[step_size*i:step_size*(i+1)]
                            loop_list[-1].index.name = new_index_name

                        # add 2D object data, all based on a unique dimension
                        # within netCDF, to loaded data dictionary
                        loadedVars[obj_key_name] = loop_list
                        del loop_list

                # prepare dataframe index for this netcdf file
                time_var = loadedVars.pop(epoch_name)

                # convert from GPS seconds to seconds used in pandas (unix time,
                # no leap)
                # time_var = convert_gps_to_unix_seconds(time_var)
                loadedVars[epoch_name] = \
                    pds.to_datetime((1E6 * time_var).astype(int))
                running_store.append(loadedVars)
                running_idx += len(loadedVars[epoch_name])

                if strict_meta:
                    if saved_mdata is None:
                        saved_mdata = copy.deepcopy(mdata)
                    elif (mdata != saved_mdata):
                        raise ValueError(' '.join(('Metadata across filenames',
                                                   'is not the same.')))

        # combine all of the data loaded across files together
        out = []
        for item in running_store:
            out.append(pds.DataFrame.from_records(item, index=epoch_name))
        out = pds.concat(out, axis=0)
    else:
        if len(fnames) == 1:
            out = xr.open_dataset(fnames[0])
        else:
            out = xr.open_mfdataset(fnames, combine='by_coords')
        for key in out.variables.keys():
            # Copy the variable attributes from the data object to the metadata
            meta_dict = {}
            for nc_key in out.variables[key].attrs.keys():
                meta_dict[nc_key] = out.variables[key].attrs[nc_key]
            mdata[key] = meta_dict
            # Remove variable attributes from the data object
            out.variables[key].attrs = {}
        # Copy the file attributes from the data object to the metadata
        for d in out.attrs.keys():
            if hasattr(mdata, d):
                mdata.__setattr__(d+'_', out.attrs[d])
            else:
                mdata.__setattr__(d, out.attrs[d])
        # Remove attributes from the data object
        out.attrs = {}

    return out, mdata



import collections
import importlib
import warnings
import numpy as np
import pandas as pds

from pysat.ssnl.avg import _calc_2d_median


class Constellation(object):
    """Manage and analyze data from multiple pysat Instruments.

    Created as part of a Spring 2018 UTDesign project.
    """
    def __init__(self, instruments=None, name=None):
        """
        Constructs a Constellation given a list of instruments or the name of
        a file with a pre-defined constellation.

        Parameters
        ----------
        instruments : list
            a list of pysat Instruments
        name : string
            Name of a file in pysat/constellations containing a list of
            instruments.

        Note
        ----
        The name and instruments parameters should not both be set.
        If neither is given, an empty constellation will be created.
        """

        if instruments and name:
            raise ValueError('When creating a constellation, please specify '
                             'a list of instruments or a name, not both.')
        elif instruments and not hasattr(instruments, '__getitem__'):
            raise ValueError('Constellation: Instruments must be list-like.')

        if instruments:
            self.instruments = instruments
        elif name:
            const = importlib.import_module('pysat.constellations.'+name)
            self.instruments = const.instruments
        else:
            self.instruments = []

    def __getitem__(self, *args, **kwargs):
        """
        Look up a member Instrument by index.
        """
        return self.instruments.__getitem__(*args, **kwargs)

    def __str__(self):
        """
        Print names of instruments within constellation.
        """
        output_str = '\npysat Constellation object:\n'

        for instr in self.instruments:
            output_str += instr.name + '\n'

        return output_str

    def set_bounds(self, start, stop):
        """
        Sets boundaries for all instruments in constellation
        """
        for instrument in self.instruments:
            instrument.bounds = (start, stop)

    def data_mod(self, *args, **kwargs):
        """
        Register a function to modify data of member Instruments.

        The function is not partially applied to modify member data.

        When the Constellation receives a function call to register
        a function for data modification, it passes the call to each
        instrument and registers it in the instrument's pysat.Custom queue.

        (Wraps pysat.Custom.add; documentation of that function is
        reproduced here.)

        Parameters
        ----------
            function : string or function object
                name of function or function object to be added to queue

            kind : {'add, 'modify', 'pass'}
                add
                    Adds data returned from fuction to instrument object.
                modify
                    pysat instrument object supplied to routine. Any and all
                    changes to object are retained.
                pass
                    A copy of pysat object is passed to function. No
                    data is accepted from return.

            at_pos : string or int
                insert at position. (default, insert at end).
            args : extra arguments

        Note
        ----
        Allowed `add` function returns:

        - {'data' : pandas Series/DataFrame/array_like,
          'units' : string/array_like of strings,
          'long_name' : string/array_like of strings,
          'name' : string/array_like of strings (iff data array_like)}

        - pandas DataFrame, names of columns are used

        - pandas Series, .name required

        - (string/list of strings, numpy array/list of arrays)
        """

        for instrument in self.instruments:
            instrument.custom.add(*args, **kwargs)

    def load(self, *args, **kwargs):
        """
        Load instrument data into instrument object.data

        (Wraps pysat.Instrument.load; documentation of that function is
        reproduced here.)

        Parameters
        ---------
        yr : integer
            Year for desired data
        doy : integer
            day of year
        data : datetime object
            date to load
        fname : 'string'
            filename to be loaded
        verifyPad : boolean
            if true, padding data not removed (debug purposes)
        """

        for instrument in self.instruments:
            instrument.load(*args, **kwargs)

    def add(self, bounds1, label1, bounds2, label2, bin3, label3,
            data_label):
        """
        Combines signals from multiple instruments within
        given bounds.

        .. deprecated:: 2.2.0
          `add` will be removed in pysat 3.0.0, it will be added to pysatSeasons

        Parameters
        ----------
        bounds1 : (min, max)
            Bounds for selecting data on the axis of label1
            Data points with label1 in [min, max) will be considered.
        label1 : string
            Data label for bounds1 to act on.
        bounds2 : (min, max)
            Bounds for selecting data on the axis of label2
            Data points with label1 in [min, max) will be considered.
        label2 : string
            Data label for bounds2 to act on.
        bin3 : (min, max, #bins)
            Min and max bounds and number of bins for third axis.
        label3 : string
            Data label for third axis.
        data_label : array of strings
            Data label(s) for data product(s) to be averaged.

        Returns
        -------
        median : dictionary
            Dictionary indexed by data label, each value of which is a
            dictionary with keys 'median', 'count', 'avg_abs_dev', and
            'bin' (the values of the bin edges.)
        """

        warnings.warn(' '.join(["constellation.add is deprecated and will be",
                                "removed in pysat 3.0.0. This functionality",
                                "will be added to pysatSeasons in the"
                                "future."]),
                      DeprecationWarning, stacklevel=2)

        # TODO Update for 2.7 compatability.
        if isinstance(data_label, str):
            data_label = [data_label, ]
        elif not isinstance(data_label, collections.Sequence):
            raise ValueError("Please pass data_label as a string or "
                             "collection of strings.")

        # Modeled after pysat.ssnl.median2D

        # Make bin boundaries.
        # y: values at label3
        # z: *data_labels
        biny = np.linspace(bin3[0], bin3[1], bin3[2]+1)

        numy = len(biny)-1
        numz = len(data_label)

        # Ranges
        yarr, zarr = map(np.arange, (numy, numz))

        # Store data here.
        ans = [[[collections.deque()] for j in yarr] for k in zarr]

        # Filter data by bounds and bin it.
        # Idiom for loading all of the data in an instrument's bounds.
        for inst in self:
            for inst in inst:
                if len(inst.data) != 0:
                    # Select indicies for each piece of data we're interest in.
                    # Not all of this data is in bounds on label3 but we'll
                    #  sort this later.
                    min1, max1 = bounds1
                    min2, max2 = bounds2
                    data1 = inst.data[label1]
                    data2 = inst.data[label2]
                    in_bounds, = np.where((min1 <= data1) & (data1 < max1) &
                                          (min2 <= data2) & (data2 < max2))
                    # Grab the data in bounds on data1, data2.
                    data_considered = inst.data.iloc[in_bounds]

                    y_indexes = np.digitize(data_considered[label3], biny) - 1

                    # Iterate over the bins along y
                    for yj in yarr:
                        # Indices of data in this bin
                        yindex, = np.where(y_indexes == yj)

                        # If there's data in this bin
                        if len(yindex) > 0:
                            ydx = data_considered.index[yindex.astype(int)]
                            # For each data label, add the points.
                            for zk in zarr:
                                dlab = data_label[zk]
                                ans[zk][yj][0].extend(
                                    data_considered.loc[ydx, dlab].tolist())

        # Now for the averaging.
        # Let's, try .. packing the answers for the 2d function.
        numx = 1
        xarr = np.arange(numx)
        binx = None

        # TODO modify output
        out_2d = _calc_2d_median(ans, data_label, binx, biny, xarr, yarr,
                                 zarr, numx, numy, numz)

        # Transform output
        output = {}
        for i, label in enumerate(data_label):
            median = [r[0] for r in out_2d[label]['median']]
            count = [r[0] for r in out_2d[label]['count']]
            dev = [r[0] for r in out_2d[label]['avg_abs_dev']]
            output[label] = {'median': median,
                             'count': count,
                             'avg_abs_dev': dev,
                             'bin': out_2d[label]['bin_y']}
        return output

    def difference(self, instrument1, instrument2, bounds, data_labels,
                   cost_function):
        """
        Calculates the difference in signals from multiple
        instruments within the given bounds.

        .. deprecated:: 2.2.0
          `difference` will be removed in pysat 3.0.0, it will be added to
          pysatSeasons

        Parameters
        ----------
        instrument1 : Instrument
            Information must already be loaded into the
            instrument.

        instrument2 : Instrument
            Information must already be loaded into the
            instrument.

        bounds : list of tuples in the form (inst1_label, inst2_label,
            min, max, max_difference)
            inst1_label are inst2_label are labels for the data in
            instrument1 and instrument2
            min and max are bounds on the data considered
            max_difference is the maximum difference between two points
            for the difference to be calculated

        data_labels : list of tuples of data labels
            The first key is used to access data in s1
            and the second data in s2.

        cost_function : function
            function that operates on two rows of the instrument data.
            used to determine the distance between two points for finding
            closest points

        Returns
        -------
        data_df: pandas DataFrame
            Each row has a point from instrument1, with the keys
            preceded by '1_', and a point within bounds on that point
            from instrument2 with the keys preceded by '2_', and the
            difference between the instruments' data for all the labels
            in data_labels

        Created as part of a Spring 2018 UTDesign project.
        """

        """
        Draft Pseudocode
        ----------------
        Check integrity of inputs.

        Let STD_LABELS be the constant tuple:
        ("time", "lat", "long", "alt")

        Note: modify so that user can override labels for time,
        lat, long, data for each satelite.

        // We only care about the data currently loaded
           into each object.

        Let start be the later of the datetime of the
         first piece of data loaded into s1, the first
         piece of data loaded into s2, and the user
         supplied start bound.

        Let end be the later of the datetime of the first
         piece of data loaded into s1, the first piece
         of data loaded into s2, and the user supplied
         end bound.

        If start is after end, raise an error.

        // Let data be the 2D array of deques holding each piece
        //  of data, sorted into bins by lat/long/alt.

        Let s1_data (resp s2_data) be data from s1.data, s2.data
        filtered by user-provided lat/long/alt bounds, time bounds
        calculated.

        Let data be a dictionary of lists with the keys
        [ dl1 for dl1, dl2 in data_labels ] +
        STD_LABELS +
        [ lb+"2" for lb in STD_LABELS ]

        For each piece of data s1_point in s1_data:

            # Hopefully np.where is very good, because this
            #  runs O(n) times.
            # We could try reusing selections, maybe, if needed.
            #  This would probably involve binning.
            Let s2_near be the data from s2.data within certain
             bounds on lat/long/alt/time using 8 statements to
             numpy.where. We can probably get those defaults from
             the user or handy constants / config?

            # We could try a different algorithm for closest pairs
            # of points.

            Let distance be the numpy array representing the
             distance between s1_point and each point in s2_near.

            # S: Difference for others: change this line.
            For each of those, calculate the spatial difference
             from the s1 using lat/long/alt. If s2_near is
             empty; break loop.

            Let s2_nearest be the point in s2_near corresponding
             to the lowest distance.

            Append to data: a point, indexed by the time from
             s1_point, containing the following data:

            # note
            Let n be the length of data["time"].
            For each key in data:
                Assert len(data[key]) == n
            End for.

            # Create data row to pass to pandas.
            Let row be an empty dict.
            For dl1, dl2 in data_labels:
                Append s1_point[dl1] - s2_nearest[dl2] to data[dl1].

            For key in STD_LABELS:
                Append s1_point[translate[key]] to data[key]
                key = key+"2"
                Append s2_nearest[translate[key]] to data[key]

        Let data_df be a pandas dataframe created from the data
        in data.

        return { 'data': data_df, 'start':start, 'end':end }
        """

        warnings.warn(' '.join(["constellation.difference is deprecated and",
                                "will be removed in pysat 3.0.0. This",
                                "functionality will be added to pysatSeasons"
                                "in the future."]),
                      DeprecationWarning, stacklevel=2)

        labels = [dl1 for dl1, dl2 in data_labels] + \
            ['1_' + b[0] for b in bounds] + ['2_' + b[1] for b in bounds] + \
            ['dist']
        data = {label: [] for label in labels}

        # Apply bounds
        inst1 = instrument1.data
        inst2 = instrument2.data
        for b in bounds:
            label1 = b[0]
            label2 = b[1]
            low = b[2]
            high = b[3]

            data1 = inst1[label1]
            ind1 = np.where((data1 >= low) & (data1 < high))
            inst1 = inst1.iloc[ind1]

            data2 = inst2[label2]
            ind2 = np.where((data2 >= low) & (data2 < high))
            inst2 = inst2.iloc[ind2]

        for i, s1_point in inst1.iterrows():
            # Gets points in instrument2 within the given bounds
            s2_near = instrument2.data
            for b in bounds:
                label1 = b[0]
                label2 = b[1]
                s1_val = s1_point[label1]
                max_dist = b[4]
                minbound = s1_val - max_dist
                maxbound = s1_val + max_dist

                data2 = s2_near[label2]
                indices = np.where((data2 >= minbound) & (data2 < maxbound))
                s2_near = s2_near.iloc[indices]

            # Finds nearest point to s1_point in s2_near
            s2_nearest = None
            min_dist = float('NaN')
            for j, s2_point in s2_near.iterrows():
                dist = cost_function(s1_point, s2_point)
                if dist < min_dist or min_dist != min_dist:
                    min_dist = dist
                    s2_nearest = s2_point

            data['dist'].append(min_dist)

            # Append difference to data dict
            for dl1, dl2 in data_labels:
                if s2_nearest is not None:
                    data[dl1].append(s1_point[dl1] - s2_nearest[dl2])
                else:
                    data[dl1].append(float('NaN'))

            # Append the rest of the row
            for b in bounds:
                label1 = b[0]
                label2 = b[1]
                data['1_' + label1].append(s1_point[label1])
                if s2_nearest is not None:
                    data['2_' + label2].append(s2_nearest[label2])
                else:
                    data['2_' + label2].append(float('NaN'))

        data_df = pds.DataFrame(data=data)
        return data_df



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------

import importlib
import pytest
import sys

import pysat


class TestWithRegistration():
    """ Test class for unit/integration tests that need registered Instruments
    """

    def setup(self):
        """ Set up the environment by registering three Instruments
        """
        # Define the modules and platforms
        self.modules = [('package1.module1', 'platname1', 'name1'),
                        ('package11.module2', 'platname1', 'name2'),
                        ('package2.module2', 'platname2', 'name2')]

        self.module_names = [mod[0] for mod in self.modules]
        self.platforms = [mod[1] for mod in self.modules]
        self.names = [mod[2] for mod in self.modules]

        # Ensure a clean starting environment, as teardown may have failed
        pysat.utils.registry.remove(self.platforms, self.names)
        self.ensure_not_in_stored_modules()

        # Create modules
        self.create_and_verify_fake_modules()

        # Set common testing attributes
        self.module_name = None
        self.platform = None
        self.name = None

        return

    def teardown(self):
        """ Tear down the environment by removing the fake modules
        """
        pysat.utils.registry.remove(self.platforms, self.names)
        self.ensure_not_in_stored_modules()

        del self.modules, self.module_names, self.platforms, self.names
        del self.module_name, self.platform, self.name

        return

    def ensure_not_in_stored_modules(self):
        """Ensure modules not in stored pysat.user_modules

        Raises
        ------
        AssertionError
            If any of the supplied Instrument modules is registered

        """
        # Get the current Instruments in registry by module
        saved_modules = pysat.utils.registry.load_saved_modules()

        # Ensure the desired modules are not yet registered
        for self.module_name, self.platform, self.name in self.modules:
            if self.platform in saved_modules:
                assert self.module_name not in saved_modules[self.platform]
            else:
                # platform not present, so not registered
                assert True
            with pytest.raises(KeyError):
                pysat.Instrument(self.platform, self.name)

        return

    def create_and_verify_fake_modules(self):
        """Create fake modules and verify instantiation of Instrument modules

        Raises
        ------
        AssertionError
            If fake Instrument platform or name is incorrect

        """

        # Create modules based on supplied module, platform, and name
        for self.module_name, self.platform, self.name in self.modules:
            # Create a fake module
            self.create_fake_module()

            # Test fake module by importing and loading it
            self.module_name = importlib.import_module(self.module_name)
            self.verify_inst_instantiation()

        return

    def create_fake_module(self):
        """Creates fake module and package from pysat_testing test Instrument
        """

        # Use pysat_testing as base instrument
        file_path = pysat.instruments.pysat_testing.__file__
        fake_package, fake_module = self.module_name.split('.')

        # Python 3.5+ implementation from https://stackoverflow.com/a/51575963
        # empty string to allow any file
        importlib.machinery.SOURCE_SUFFIXES.append('')
        spec = importlib.util.spec_from_file_location(fake_module,
                                                      file_path)
        inst = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(inst)

        # Update the platform and name
        inst.platform = self.platform
        inst.name = self.name

        package_spec = importlib.util.spec_from_loader(fake_package, None,
                                                       is_package=True)
        package = importlib.util.module_from_spec(package_spec)

        setattr(package, fake_module, inst)
        setattr(package, '__all__', [fake_module])
        sys.modules[fake_package] = package
        sys.modules['.'.join([fake_package, fake_module])] = inst

        return

    def verify_inst_instantiation(self):
        """Verify that information is sufficient for importing module

        Note
        ----
        Checks for presence in pysat.user_modules

        Raises
        ------
        AssertionError
            If fake Instrument module or platform/name combo is incorrect

        """

        if self.module_name is None:
            # Load Instrument by platform and name
            inst = pysat.Instrument(self.platform, self.name)
        else:
            # Load Instrument by module
            inst = pysat.Instrument(inst_module=self.module_name)

        # ensure we have the correct one
        assert inst.platform == self.platform
        assert inst.name == self.name

        return



import copy
from io import StringIO
import logging
import pytest

import pysat


def mult_data(inst, mult, dkey="mlt"):
    """Function to add a multiplied data value to an Instrument

    Parameters
    ----------
    inst : pysat.Instrument
        pysat Instrument object
    mult : float or int
        Multiplicative value
    dkey : str
        Key for data that will be multiplied (default='mlt')

    Note
    ----
    Adds multiplied data to Instrument using a key name with the format
    {mult}x{dkey}

    """
    # Construct the new data key
    out_key = '{:.0f}x{:s}'.format(mult, dkey)

    # Add the new data to the instrument
    inst.data[out_key] = mult * inst.data[dkey]

    return


class TestLogging():
    def setup(self):
        """Runs before every method to create a clean testing setup.
        """
        self.testInst = pysat.Instrument('pysat', 'testing', num_samples=10,
                                         clean_level='clean',
                                         update_files=False)
        self.out = ''
        self.log_capture = StringIO()
        pysat.logger.addHandler(logging.StreamHandler(self.log_capture))
        pysat.logger.setLevel(logging.WARNING)

    def teardown(self):
        """Runs after every method to clean up previous testing.
        """
        del self.testInst, self.out, self.log_capture

    def test_custom_pos_warning(self):
        """Test for logging warning if inappropriate position specified
        """

        self.testInst.custom_attach(lambda inst: inst.data['mlt'] * 2.0,
                                    at_pos=3)
        self.out = self.log_capture.getvalue()

        assert self.out.find(
            "unknown position specified, including function at end") >= 0


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup.
        """
        self.testInst = pysat.Instrument('pysat', 'testing', num_samples=10,
                                         clean_level='clean',
                                         update_files=True)
        self.load_date = pysat.instruments.pysat_testing._test_dates['']['']
        self.testInst.load(date=self.load_date)
        self.custom_args = [2]
        self.out = None

    def teardown(self):
        """Runs after every method to clean up previous testing.
        """
        del self.testInst, self.out, self.custom_args

    def test_basic_str(self):
        """Check for lines from each decision point in str"""
        self.out = self.testInst.__str__()
        assert isinstance(self.out, str)

        # No custom functions
        assert self.out.find('0 applied') > 0

    def test_basic_repr(self):
        """Test __repr__ with a custom method"""

        self.testInst.custom_attach(mult_data, args=self.custom_args)
        self.testInst.custom_attach(mult_data, args=self.custom_args)
        self.out = self.testInst.__repr__()
        assert isinstance(self.out, str)
        assert self.out.find("'function'") >= 0

    def test_basic_str_w_function(self):
        """Check for lines from each decision point in str"""

        self.testInst.custom_attach(mult_data, args=self.custom_args,
                                    kwargs={'dkey': 'mlt'})

        # Execute custom method
        self.testInst.load(date=self.load_date)

        # Store output to be tested
        self.out = self.testInst.__str__()
        assert isinstance(self.out, str)

        # No custom functions
        assert self.out.find('1 applied') > 0
        assert self.out.find('mult_data') > 0
        assert self.out.find('Args') > 0
        assert self.out.find('Kwargs') > 0

    def test_single_modifying_custom_function_error(self):
        """Test for error when custom function loaded as modify returns a value
        """
        def custom_with_return_data(inst):
            inst.data['doubleMLT'] = 2.0 * inst.data.mlt
            return 5.0 * inst.data['mlt']

        self.testInst.custom_attach(custom_with_return_data)
        with pytest.raises(ValueError) as verr:
            self.testInst.load(date=self.load_date)

        estr = 'Custom functions should not return any information via return'
        assert str(verr).find(estr) >= 0

    def test_custom_keyword_instantiation(self):
        """Test adding custom methods at Instrument instantiation
        """

        self.testInst.custom_attach(mult_data, args=self.custom_args,
                                    kwargs={'dkey': 'mlt'})
        self.testInst.custom_attach(mult_data, args=self.custom_args)

        # Create another instance of pysat.Instrument and add custom
        # via the input keyword
        custom = [{'function': mult_data, 'args': self.custom_args,
                   'kwargs': {'dkey': 'mlt'}},
                  {'function': mult_data, 'args': self.custom_args}]
        testInst2 = pysat.Instrument('pysat', 'testing', custom=custom)

        # Ensure both instruments have the same custom_* attributes
        assert self.testInst.custom_functions == testInst2.custom_functions
        assert self.testInst.custom_args == testInst2.custom_args
        assert self.testInst.custom_kwargs == testInst2.custom_kwargs

    def test_custom_positioning(self):
        """Test custom method ordering specification
        """

        self.testInst.custom_attach(mult_data, args=[3],
                                    kwargs={'dkey': '2xmlt'})
        self.testInst.custom_attach(mult_data, at_pos=0, args=self.custom_args)

        # Create another instance of pysat.Instrument and add custom
        # via the input keyword
        custom = [{'function': mult_data, 'args': [3],
                   'kwargs': {'dkey': '2xmlt'}},
                  {'function': mult_data, 'at_pos': 0,
                   'args': self.custom_args}]
        testInst2 = pysat.Instrument('pysat', 'testing', custom=custom)

        # Ensure both Instruments have the same custom_* attributes
        assert self.testInst.custom_functions == testInst2.custom_functions
        assert self.testInst.custom_args == testInst2.custom_args
        assert self.testInst.custom_kwargs == testInst2.custom_kwargs

        # Ensure the run order was correct
        assert self.testInst.custom_args[0] == self.custom_args
        assert self.testInst.custom_args[1] == [3]

    def test_custom_keyword_instantiation_poor_format(self):
        """Test for error when custom missing keywords at instantiation
        """
        req_words = ['function']
        real_custom = [{'function': 1, 'args': [0, 1],
                        'kwargs': {'kwarg1': True, 'kwarg2': False}}]
        for i, word in enumerate(req_words):
            custom = copy.deepcopy(real_custom)
            custom[0].pop(word)

            # Ensure that any of the missing required words raises an error
            with pytest.raises(ValueError) as err:
                pysat.Instrument('pysat', 'testing', custom=custom)

            # Ensure correct error for the missing parameter (word)
            assert str(err).find(word) >= 0
            assert str(err).find('Input dict to custom is missing') >= 0

        return

    def test_clear_functions(self):
        """Test successful clearance of custom functions
        """
        self.testInst.custom_attach(lambda inst, imult, out_units='h':
                                    {'data': (inst.data.mlt * imult).values,
                                     'long_name': 'doubleMLTlong',
                                     'units': out_units, 'name': 'doubleMLT'},
                                    args=[2], kwargs={"out_units": "hours1"})

        # Test to see that the custom function was attached
        assert len(self.testInst.custom_functions) == 1
        assert len(self.testInst.custom_args) == 1
        assert len(self.testInst.custom_kwargs) == 1

        self.testInst.custom_clear()
        # Test to see that the custom function was cleared
        assert self.testInst.custom_functions == []
        assert self.testInst.custom_args == []
        assert self.testInst.custom_kwargs == []


# Repeat the above tests with xarray
class TestBasicsXarray(TestBasics):
    def setup(self):
        """Runs before every method to create a clean testing setup.
        """
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         num_samples=10, clean_level='clean')
        self.load_date = pysat.instruments.pysat_testing_xarray._test_dates
        self.load_date = self.load_date['']['']
        self.testInst.load(date=self.load_date)
        self.custom_args = [2]

    def teardown(self):
        """Runs after every method to clean up previous testing.
        """
        del self.testInst, self.load_date, self.custom_args


class TestConstellationBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup
        """
        self.testConst = pysat.Constellation(instruments=[
            pysat.Instrument('pysat', 'testing', num_samples=10,
                             clean_level='clean',
                             update_files=True)
            for i in range(5)])
        self.load_date = pysat.instruments.pysat_testing._test_dates['']['']
        self.testConst.load(date=self.load_date)
        self.custom_args = [2]

    def teardown(self):
        """ Runs after every method to clean up previous testing
        """
        del self.testConst, self.load_date, self.custom_args

    def test_basic_repr(self):
        """Test __repr__ with a custom method"""

        self.testConst.custom_attach(mult_data, args=self.custom_args)
        self.out = self.testConst.__repr__()
        assert isinstance(self.out, str)
        assert self.out.find("'function'") >= 0

    def test_single_modifying_custom_function_error(self):
        """Test for error when custom function loaded as modify returns a value
        """
        def custom_with_return_data(inst):
            inst.data['doubleMLT'] = 2.0 * inst.data.mlt
            return 5.0 * inst.data['mlt']

        self.testConst.custom_attach(custom_with_return_data)
        with pytest.raises(ValueError) as verr:
            self.testConst.load(date=self.load_date)

        estr = 'Custom functions should not return any information via return'
        assert str(verr).find(estr) >= 0

    def test_custom_keyword_instantiation(self):
        """Test adding custom methods at Instrument instantiation
        """

        self.testConst.custom_attach(mult_data, args=self.custom_args,
                                     kwargs={'dkey': 'mlt'})
        self.testConst.custom_attach(mult_data, args=self.custom_args)

        # Create another instance of pysat.Instrument and add custom
        # via the input keyword
        custom = [{'function': mult_data, 'args': self.custom_args,
                   'kwargs': {'dkey': 'mlt'}},
                  {'function': mult_data, 'args': self.custom_args}]
        testConst2 = pysat.Constellation(instruments=[
            pysat.Instrument('pysat', 'testing', num_samples=10,
                             clean_level='clean', custom=custom)
            for i in range(5)])

        # Ensure all instruments within both constellations have the same
        # custom_* attributes
        for i, inst in enumerate(self.testConst.instruments):
            inst2 = testConst2.instruments[i]
            assert inst.custom_functions == inst2.custom_functions
            assert inst.custom_args == inst2.custom_args
            assert inst.custom_kwargs == inst2.custom_kwargs

    def test_clear_functions(self):
        """Test successful clearance of custom functions
        """
        self.testConst.custom_attach(lambda inst, imult, out_units='h':
                                     {'data': (inst.data.mlt * imult).values,
                                      'long_name': 'doubleMLTlong',
                                      'units': out_units, 'name': 'doubleMLT'},
                                     args=[2], kwargs={"out_units": "hours1"})

        # Test to see that the custom function was attached
        for inst in self.testConst.instruments:
            assert len(inst.custom_functions) == 1
            assert len(inst.custom_args) == 1
            assert len(inst.custom_kwargs) == 1

        # Test to see that the custom function was cleared
        self.testConst.custom_clear()
        for inst in self.testConst.instruments:
            assert inst.custom_functions == []
            assert inst.custom_args == []
            assert inst.custom_kwargs == []



"""
tests the pysat meta object and code
"""
import glob
import numpy as np
import os
import sys

from nose.tools import raises
import pandas as pds
import tempfile

import pysat
import pysat.instruments.pysat_testing

if sys.version_info[0] >= 3:
    from importlib import reload as re_load
else:
    re_load = reload


def create_dir(inst=None, temporary_file_list=False):
    if inst is None:
        # create instrument
        inst = pysat.Instrument(platform='pysat', name='testing',
                                temporary_file_list=temporary_file_list)

    # create data directories
    try:
        os.makedirs(inst.files.data_path)
    except OSError:
        pass
    return


def remove_files(inst=None):
    # remove any files
    temp_dir = inst.files.data_path
    for the_file in os.listdir(temp_dir):
        if (the_file[0:13] == 'pysat_testing') & \
                (the_file[-19:] == '.pysat_testing_file'):
            file_path = os.path.join(temp_dir, the_file)
            if os.path.isfile(file_path):
                os.unlink(file_path)


# create year doy file set
def create_files(inst, start, stop, freq=None, use_doy=True, root_fname=None,
                 content=None):

    if freq is None:
        freq = '1D'
    dates = pysat.utils.time.create_date_range(start, stop, freq=freq)

    if root_fname is None:
        root_fname = ''.join(('pysat_testing_junk_{year:04d}_gold_{day:03d}_',
                              'stuff.pysat_testing_file'))
    # create empty file
    for date in dates:
        yr, doy = pysat.utils.time.getyrdoy(date)
        if use_doy:
            doy = doy
        else:
            doy = date.day

        fname = os.path.join(inst.files.data_path, root_fname.format(year=yr,
                             day=doy, month=date.month, hour=date.hour,
                             minute=date.minute, second=date.second))
        with open(fname, 'w') as f:
            if content is not None:
                f.write(content)


def list_files(tag=None, sat_id=None, data_path=None, format_str=None):
    """Return a Pandas Series of every file for chosen satellite data"""

    if format_str is None:
        format_str = ''.join(('pysat_testing_junk_{year:04d}_gold_{day:03d}_',
                              'stuff_{month:02d}_{hour:02d}_{minute:02d}_',
                              '{second:02d}.pysat_testing_file'))

    if tag is not None:
        if tag == '':
            return pysat.Files.from_os(data_path=data_path,
                                       format_str=format_str)
        else:
            raise ValueError('Unrecognized tag name')
    else:
        raise ValueError('A tag name must be passed ')


class TestNoDataDir():

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.temporary_file_list = False
        # store current pysat directory
        self.saved_data_path = pysat.data_dir

        pysat.data_dir = ''
        re_load(pysat._files)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        pysat.data_dir = self.saved_data_path
        re_load(pysat._files)

    @raises(Exception)
    def test_no_data_dir(self):
        _ = pysat.Instrument()


class TestBasics():

    temporary_file_list = False

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # store current pysat directory
        self.data_path = pysat.data_dir

        # create temporary directory
        dir_name = tempfile.mkdtemp()
        pysat.utils.set_data_dir(dir_name, store=False)

        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             temporary_file_list=self.temporary_file_list)
        # create testing directory
        create_dir(self.testInst)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        remove_files(self.testInst)
        try:
            pysat.utils.set_data_dir(self.data_path, store=False)
        except:
            pass
        del self.testInst

    def test_parse_delimited_filename(self):
        """Check ability to parse delimited files"""
        # Note: Can be removed if future instrument that uses delimited
        # filenames is added to routine travis end-to-end testing
        fname = ''.join(('test_{year:4d}_{month:2d}_{day:2d}_{hour:2d}',
                         '_{minute:2d}_{second:2d}_{version:2s}_r02.cdf'))
        year = np.ones(6)*2009
        month = np.ones(6)*12
        day = np.array([12, 15, 17, 19, 22, 24])
        hour = np.array([8, 10, 6, 18, 3, 23])
        minute = np.array([8, 10, 6, 18, 3, 59])
        second = np.array([58, 11, 26, 2, 18, 59])
        version = np.array(['v1', 'v2', 'r1', 'r3', 'v5', 'a6'])
        file_list = []
        for i in range(6):
            file_list.append(fname.format(year=year[i].astype(int),
                                          month=month[i].astype(int),
                                          day=day[i], hour=hour[i],
                                          minute=minute[i], second=second[i],
                                          version=version[i]))

        file_dict = pysat._files.parse_delimited_filenames(file_list, fname,
                                                           '_')
        assert np.all(file_dict['year'] == year)
        assert np.all(file_dict['month'] == month)
        assert np.all(file_dict['day'] == day)
        assert np.all(file_dict['hour'] == hour)
        assert np.all(file_dict['minute'] == minute)
        assert np.all(file_dict['day'] == day)
        assert np.all(file_dict['version'] == version)
        assert (file_dict['revision'] is None)

    def test_year_doy_files_direct_call_to_from_os(self):
        # create a bunch of files by year and doy
        start = pysat.datetime(2008, 1, 1)
        stop = pysat.datetime(2009, 12, 31)
        create_files(self.testInst, start, stop, freq='1D')
        # use from_os function to get pandas Series of files and dates
        files = pysat.Files.from_os(data_path=self.testInst.files.data_path,
                                    format_str=''.join(('pysat_testing_junk_',
                                                        '{year:04d}_gold_',
                                                        '{day:03d}_stuff.',
                                                        'pysat_testing_file')))
        # check overall length
        check1 = len(files) == (365 + 366)
        # check specific dates
        check2 = pds.to_datetime(files.index[0]) == pysat.datetime(2008, 1, 1)
        check3 = pds.to_datetime(files.index[365]) == \
            pysat.datetime(2008, 12, 31)
        check4 = pds.to_datetime(files.index[-1]) == \
            pysat.datetime(2009, 12, 31)
        assert(check1 & check2 & check3 & check4)

    def test_year_doy_files_no_gap_in_name_direct_call_to_from_os(self):
        # create a bunch of files by year and doy
        start = pysat.datetime(2008, 1, 1)
        stop = pysat.datetime(2009, 12, 31)
        create_files(self.testInst, start, stop, freq='1D',
                     root_fname=''.join(('pysat_testing_junk_{year:04d}',
                                         '{day:03d}_stuff.pysat_testing_',
                                         'file')))
        # use from_os function to get pandas Series of files and dates
        files = pysat.Files.from_os(data_path=self.testInst.files.data_path,
                                    format_str=''.join(('pysat_testing_junk_',
                                                        '{year:04d}{day:03d}_',
                                                        'stuff.pysat_testing_',
                                                        'file')))
        # check overall length
        check1 = len(files) == (365 + 366)
        # check specific dates
        check2 = pds.to_datetime(files.index[0]) == pysat.datetime(2008, 1, 1)
        check3 = pds.to_datetime(files.index[365]) == \
            pysat.datetime(2008, 12, 31)
        check4 = pds.to_datetime(files.index[-1]) == \
            pysat.datetime(2009, 12, 31)
        assert(check1 & check2 & check3 & check4)

    def test_year_month_day_files_direct_call_to_from_os(self):
        # create a bunch of files by year and doy
        start = pysat.datetime(2008, 1, 1)
        stop = pysat.datetime(2009, 12, 31)
        create_files(self.testInst, start, stop, freq='1D', use_doy=False,
                     root_fname=''.join(('pysat_testing_junk_{year:04d}_gold_',
                                         '{day:03d}_stuff_{month:02d}.pysat_',
                                         'testing_file')))
        # use from_os function to get pandas Series of files and dates
        files = pysat.Files.from_os(data_path=self.testInst.files.data_path,
                                    format_str=''.join(('pysat_testing_junk_',
                                                        '{year:04d}_gold_',
                                                        '{day:03d}_stuff_',
                                                        '{month:02d}.pysat_',
                                                        'testing_file')))
        # check overall length
        check1 = len(files) == (365 + 366)
        # check specific dates
        check2 = pds.to_datetime(files.index[0]) == pysat.datetime(2008, 1, 1)
        check3 = pds.to_datetime(files.index[365]) == \
            pysat.datetime(2008, 12, 31)
        check4 = pds.to_datetime(files.index[-1]) == \
            pysat.datetime(2009, 12, 31)
        assert(check1 & check2 & check3 & check4)

    def test_year_month_day_hour_files_direct_call_to_from_os(self):
        # create a bunch of files by year and doy
        start = pysat.datetime(2008, 1, 1)
        stop = pysat.datetime(2009, 12, 31)
        create_files(self.testInst, start, stop, freq='6h',
                     use_doy=False,
                     root_fname=''.join(('pysat_testing_junk_{year:04d}_gold_',
                                         '{day:03d}_stuff_{month:02d}_',
                                         '{hour:02d}.pysat_testing_file')))
        # use from_os function to get pandas Series of files and dates
        files = pysat.Files.from_os(data_path=self.testInst.files.data_path,
                                    format_str=''.join(('pysat_testing_junk_',
                                                        '{year:04d}_gold_',
                                                        '{day:03d}_stuff_',
                                                        '{month:02d}_',
                                                        '{hour:02d}.pysat_',
                                                        'testing_file')))
        # check overall length
        check1 = len(files) == (365+366)*4-3
        # check specific dates
        check2 = pds.to_datetime(files.index[0]) == pysat.datetime(2008, 1, 1)
        check3 = pds.to_datetime(files.index[1460]) == \
            pysat.datetime(2008, 12, 31)
        check4 = pds.to_datetime(files.index[-1]) == \
            pysat.datetime(2009, 12, 31)
        assert(check1 & check2 & check3 & check4)

    def test_year_month_day_hour_minute_files_direct_call_to_from_os(self):
        root_fname = ''.join(('pysat_testing_junk_{year:04d}_gold_{day:03d}_',
                              'stuff_{month:02d}_{hour:02d}{minute:02d}.',
                              'pysat_testing_file'))
        # create a bunch of files by year and doy
        start = pysat.datetime(2008, 1, 1)
        stop = pysat.datetime(2008, 1, 4)
        create_files(self.testInst, start, stop, freq='30min',
                     use_doy=False,
                     root_fname=root_fname)
        # use from_os function to get pandas Series of files and dates
        files = pysat.Files.from_os(data_path=self.testInst.files.data_path,
                                    format_str=root_fname)
        # check overall length
        check1 = len(files) == 145
        # check specific dates
        check2 = pds.to_datetime(files.index[0]) == pysat.datetime(2008, 1, 1)
        check3 = pds.to_datetime(files.index[1]) == \
            pysat.datetime(2008, 1, 1, 0, 30)
        check4 = pds.to_datetime(files.index[10]) == \
            pysat.datetime(2008, 1, 1, 5, 0)
        check5 = pds.to_datetime(files.index[-1]) == pysat.datetime(2008, 1, 4)
        assert(check1 & check2 & check3 & check4 & check5)

    def test_year_month_day_hour_minute_second_files_direct_call_to_from_os(self):
        root_fname = ''.join(('pysat_testing_junk_{year:04d}_gold_{day:03d}_',
                              'stuff_{month:02d}_{hour:02d}_{minute:02d}_',
                              '{second:02d}.pysat_testing_file'))
        # create a bunch of files by year and doy
        start = pysat.datetime(2008, 1, 1)
        stop = pysat.datetime(2008, 1, 3)
        create_files(self.testInst, start, stop, freq='30s',
                     use_doy=False, root_fname=root_fname)
        # use from_os function to get pandas Series of files and dates
        files = pysat.Files.from_os(data_path=self.testInst.files.data_path,
                                    format_str=root_fname)
        # check overall length
        check1 = len(files) == 5761
        # check specific dates
        check2 = pds.to_datetime(files.index[0]) == pysat.datetime(2008, 1, 1)
        check3 = pds.to_datetime(files.index[1]) == \
            pysat.datetime(2008, 1, 1, 0, 0, 30)
        check4 = pds.to_datetime(files.index[-1]) == \
            pysat.datetime(2008, 1, 3)
        assert(check1 & check2 & check3 & check4)

    def test_year_month_files_direct_call_to_from_os(self):
        # create a bunch of files by year and doy
        start = pysat.datetime(2008, 1, 1)
        stop = pysat.datetime(2009, 12, 31)
        create_files(self.testInst, start, stop, freq='1MS',
                     root_fname=''.join(('pysat_testing_junk_{year:04d}_gold_',
                                         'stuff_{month:02d}.pysat_testing_',
                                         'file')))
        # use from_os function to get pandas Series of files and dates
        files = pysat.Files.from_os(data_path=self.testInst.files.data_path,
                                    format_str=''.join(('pysat_testing_junk_',
                                                        '{year:04d}_gold_',
                                                        'stuff_{month:02d}.',
                                                        'pysat_testing_file')))
        # check overall length
        check1 = len(files) == 24
        # check specific dates
        check2 = pds.to_datetime(files.index[0]) == pysat.datetime(2008, 1, 1)
        check3 = pds.to_datetime(files.index[11]) == \
            pysat.datetime(2008, 12, 1)
        check4 = pds.to_datetime(files.index[-1]) == \
            pysat.datetime(2009, 12, 1)
        assert(check1 & check2 & check3 & check4)

    def test_instrument_has_no_files(self):
        import pysat.instruments.pysat_testing

        pysat.instruments.pysat_testing.list_files = list_files
        inst = pysat.Instrument(platform='pysat', name='testing',
                                update_files=True)
        re_load(pysat.instruments.pysat_testing)
        assert(inst.files.files.empty)

    def test_instrument_has_files(self):
        import pysat.instruments.pysat_testing

        root_fname = ''.join(('pysat_testing_junk_{year:04d}_gold_{day:03d}_'
                              'stuff_{month:02d}_{hour:02d}_{minute:02d}_'
                              '{second:02d}.pysat_testing_file'))
        # create a bunch of files by year and doy
        start = pysat.datetime(2007, 12, 31)
        stop = pysat.datetime(2008, 1, 10)
        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False, root_fname=root_fname)
        # create the same range of dates
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        pysat.instruments.pysat_testing.list_files = list_files
        inst = pysat.Instrument(platform='pysat', name='testing',
                                update_files=True)
        re_load(pysat.instruments.pysat_testing)
        assert (np.all(inst.files.files.index == dates))


class TestBasicsNoFileListStorage(TestBasics):

    temporary_file_list = True


class TestInstrumentWithFiles():

    temporary_file_list = False

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # store current pysat directory
        self.data_path = pysat.data_dir
        # create temporary directory
        dir_name = tempfile.mkdtemp()
        pysat.utils.set_data_dir(dir_name, store=False)
        # create testing directory
        create_dir(temporary_file_list=self.temporary_file_list)

        # create a test instrument, make sure it is getting files from
        # filesystem
        re_load(pysat.instruments.pysat_testing)
        pysat.instruments.pysat_testing.list_files = list_files
        # create a bunch of files by year and doy
        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             temporary_file_list=self.temporary_file_list)

        self.root_fname = ''.join(('pysat_testing_junk_{year:04d}_gold_',
                                   '{day:03d}_stuff_{month:02d}_{hour:02d}_',
                                   '{minute:02d}_{second:02d}.pysat_testing_',
                                   'file'))
        start = pysat.datetime(2007, 12, 31)
        stop = pysat.datetime(2008, 1, 10)
        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False, root_fname=self.root_fname)

        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        remove_files(self.testInst)
        del self.testInst
        re_load(pysat.instruments.pysat_testing)
        re_load(pysat.instruments)
        # make sure everything about instrument state is restored
        # restore original file list, no files
        pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                         clean_level='clean',
                         update_files=True,
                         temporary_file_list=self.temporary_file_list)
        pysat.utils.set_data_dir(self.data_path, store=False)

    def test_refresh(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 10)
        stop = pysat.datetime(2008, 1, 12)

        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False,
                     root_fname=self.root_fname)
        start = pysat.datetime(2007, 12, 31)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        self.testInst.files.refresh()
        assert (np.all(self.testInst.files.files.index == dates))

    def test_refresh_on_ignore_empty_files(self):
        # setup created empty files - make sure such files can be ignored
        self.testInst.files.ignore_empty_files = True
        self.testInst.files.refresh()
        assert len(self.testInst.files.files) == 0

        # create new files with content and make sure they are captured
        start = pysat.datetime(2007, 12, 31)
        stop = pysat.datetime(2008, 1, 10)
        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False,
                     root_fname=self.root_fname,
                     content='test')
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        self.testInst.files.refresh()
        assert (np.all(self.testInst.files.files.index == dates))

    def test_instrument_with_ignore_empty_files(self):
        """Make sure new instruments can ignore empty files"""
        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             update_files=True,
                             temporary_file_list=self.temporary_file_list,
                             ignore_empty_files=True)

        assert len(self.testInst.files.files) == 0

        # create new files with content and make sure they are captured
        start = pysat.datetime(2007, 12, 31)
        stop = pysat.datetime(2008, 1, 10)
        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False,
                     root_fname=self.root_fname,
                     content='test')
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        self.testInst.files.refresh()
        assert (np.all(self.testInst.files.files.index == dates))

    def test_refresh_on_unchanged_files(self):
        start = pysat.datetime(2007, 12, 31)
        stop = pysat.datetime(2008, 1, 10)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        self.testInst.files.refresh()
        assert (np.all(self.testInst.files.files.index == dates))

    def test_get_new_files_after_refresh(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 12)

        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False,
                     root_fname=self.root_fname)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        self.testInst.files.refresh()
        new_files = self.testInst.files.get_new()

        assert (np.all(new_files.index == dates))

    def test_get_new_files_after_multiple_refreshes(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 12)

        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False,
                     root_fname=self.root_fname)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        self.testInst.files.refresh()
        self.testInst.files.refresh()
        self.testInst.files.refresh()
        new_files = self.testInst.files.get_new()
        assert (np.all(new_files.index == dates))

    def test_get_new_files_after_adding_files(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 12)

        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False,
                     root_fname=self.root_fname)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        new_files = self.testInst.files.get_new()
        assert (np.all(new_files.index == dates))

    def test_get_new_files_after_adding_files_and_adding_file(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 12)

        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False,
                     root_fname=self.root_fname)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        new_files = self.testInst.files.get_new()

        start = pysat.datetime(2008, 1, 15)
        stop = pysat.datetime(2008, 1, 18)

        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False,
                     root_fname=self.root_fname)
        dates2 = pysat.utils.time.create_date_range(start, stop, freq='100min')
        new_files2 = self.testInst.files.get_new()
        assert (np.all(new_files.index == dates) &
                np.all(new_files2.index == dates2))

    def test_get_new_files_after_deleting_files_and_adding_files(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 12)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')

        # remove files, same number as will be added
        to_be_removed = len(dates)
        for the_file in os.listdir(self.testInst.files.data_path):
            if (the_file[0:13] == 'pysat_testing') & \
                    (the_file[-19:] == '.pysat_testing_file'):
                file_path = os.path.join(self.testInst.files.data_path,
                                         the_file)
                if os.path.isfile(file_path) & (to_be_removed > 0):
                    # print(file_path)
                    to_be_removed -= 1
                    os.unlink(file_path)
        # add new files
        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False, root_fname=self.root_fname)
        # get new files
        new_files = self.testInst.files.get_new()

        assert (np.all(new_files.index == dates))

    def test_files_non_standard_pysat_directory(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 15)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')

        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             sat_id='hello',
                             directory_format='pysat_testing_{tag}_{sat_id}',
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)
        # add new files
        create_dir(self.testInst)
        remove_files(self.testInst)
        create_files(self.testInst, start, stop, freq='100min',
                     use_doy=False, root_fname=self.root_fname)

        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             sat_id='hello',
                             directory_format=''.join(('pysat_testing_',
                                                       '{tag}_{sat_id}')),
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)

        # get new files
        new_files = self.testInst.files.get_new()
        assert (np.all(self.testInst.files.files.index == dates) &
                np.all(new_files.index == dates))

    def test_files_non_standard_file_format_template(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 15)
        dates = pysat.utils.time.create_date_range(start, stop, freq='1D')

        # clear out old files, create new ones
        remove_files(self.testInst)
        create_files(self.testInst, start, stop, freq='1D',
                     use_doy=False,
                     root_fname=''.join(('pysat_testing_unique_junk_',
                                         '{year:04d}_gold_{day:03d}_stuff',
                                         '.pysat_testing_file')))

        pysat.instruments.pysat_testing.list_files = list_files
        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             file_format=''.join(('pysat_testing_unique_',
                                                  'junk_{year:04d}_gold_',
                                                  '{day:03d}_stuff',
                                                  '.pysat_testing_file')),
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)

        assert (np.all(self.testInst.files.files.index == dates))

    @raises(ValueError)
    def test_files_non_standard_file_format_template_misformatted(self):

        pysat.instruments.pysat_testing.list_files = list_files
        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             file_format=''.join(('pysat_testing_unique_',
                                                  'junk_stuff.pysat_testing',
                                                  '_file')),
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)

    @raises(ValueError)
    def test_files_non_standard_file_format_template_misformatted_2(self):

        pysat.instruments.pysat_testing.list_files = list_files
        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             file_format=15,
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)


class TestInstrumentWithFilesNoFileListStorage(TestInstrumentWithFiles):

    temporary_file_list = True


# create year doy file set with multiple versions
def create_versioned_files(inst, start=None, stop=None, freq='1D',
                           use_doy=True, root_fname=None):
    # create a bunch of files
    if start is None:
        start = pysat.datetime(2009, 1, 1)
    if stop is None:
        stop = pysat.datetime(2013, 12, 31)
    dates = pysat.utils.time.create_date_range(start, stop, freq=freq)

    versions = np.array([1, 2])
    revisions = np.array([0, 1])

    if root_fname is None:
        root_fname = ''.join(('pysat_testing_junk_{year:04d}_{month:02d}_',
                              '{day:03d}{hour:02d}{minute:02d}{second:02d}_',
                              'stuff_{version:02d}_{revision:03d}.pysat_',
                              'testing_file'))
    # create empty file
    for date in dates:
        for version in versions:
            for revision in revisions:
                yr, doy = pysat.utils.time.getyrdoy(date)
                if use_doy:
                    doy = doy
                else:
                    doy = date.day

                fname = os.path.join(inst.files.data_path,
                                     root_fname.format(year=yr,
                                                       day=doy,
                                                       month=date.month,
                                                       hour=date.hour,
                                                       minute=date.minute,
                                                       second=date.second,
                                                       version=version,
                                                       revision=revision))
                with open(fname, 'w') as f:
                    pass


def list_versioned_files(tag=None, sat_id=None, data_path=None,
                         format_str=None):
    """Return a Pandas Series of every file for chosen satellite data"""

    if format_str is None:
        format_str = ''.join(('pysat_testing_junk_{year:04d}_{month:02d}_',
                              '{day:03d}{hour:02d}{minute:02d}{second:02d}_',
                              'stuff_{version:02d}_{revision:03d}.pysat_',
                              'testing_file'))
    if tag is not None:
        if tag == '':
            return pysat.Files.from_os(data_path=data_path,
                                       format_str=format_str)
        else:
            raise ValueError('Unrecognized tag name')
    else:
        raise ValueError('A tag name must be passed ')


class TestInstrumentWithVersionedFiles():

    temporary_file_list = False

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # store current pysat directory
        self.data_path = pysat.data_dir
        # create temporary directory
        dir_name = tempfile.gettempdir()
        pysat.utils.set_data_dir(dir_name, store=False)
        # create testing directory
        create_dir(temporary_file_list=self.temporary_file_list)

        # create a test instrument, make sure it is getting files from
        # filesystem
        re_load(pysat.instruments.pysat_testing)
        pysat.instruments.pysat_testing.list_files = list_versioned_files
        # create a bunch of files by year and doy
        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             temporary_file_list=self.temporary_file_list)

        self.root_fname = ''.join(('pysat_testing_junk_{year:04d}_{month:02d}',
                                   '_{day:03d}{hour:02d}{minute:02d}',
                                   '{second:02d}_stuff_{version:02d}_',
                                   '{revision:03d}.pysat_testing_file'))
        start = pysat.datetime(2007, 12, 31)
        stop = pysat.datetime(2008, 1, 10)
        create_versioned_files(self.testInst, start, stop, freq='100min',
                               use_doy=False, root_fname=self.root_fname)

        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        remove_files(self.testInst)
        del self.testInst
        re_load(pysat.instruments.pysat_testing)
        re_load(pysat.instruments)
        # make sure everything about instrument state is restored
        # restore original file list, no files
        pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                         clean_level='clean',
                         update_files=True,
                         temporary_file_list=self.temporary_file_list)
        pysat.utils.set_data_dir(self.data_path, store=False)

    def test_refresh(self):
        # create new files and make sure that new files are captured
        # files slready exist from 2007, 12, 31 through to 10th
        start = pysat.datetime(2008, 1, 10)
        stop = pysat.datetime(2008, 1, 12)
        create_versioned_files(self.testInst, start, stop, freq='100min',
                               use_doy=False,
                               root_fname=self.root_fname)
        # create list of dates for all files that should be there
        start = pysat.datetime(2007, 12, 31)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        # update instrument file list
        self.testInst.files.refresh()
        assert (np.all(self.testInst.files.files.index == dates))

    def test_refresh_on_unchanged_files(self):

        start = pysat.datetime(2007, 12, 31)
        stop = pysat.datetime(2008, 1, 10)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        self.testInst.files.refresh()
        assert (np.all(self.testInst.files.files.index == dates))

    def test_get_new_files_after_refresh(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 12)

        create_versioned_files(self.testInst, start, stop, freq='100min',
                               use_doy=False,
                               root_fname=self.root_fname)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        self.testInst.files.refresh()
        new_files = self.testInst.files.get_new()

        assert (np.all(new_files.index == dates))

    def test_get_new_files_after_multiple_refreshes(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 12)

        create_versioned_files(self.testInst, start, stop, freq='100min',
                               use_doy=False,
                               root_fname=self.root_fname)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        self.testInst.files.refresh()
        self.testInst.files.refresh()
        self.testInst.files.refresh()
        new_files = self.testInst.files.get_new()

        assert (np.all(new_files.index == dates))

    def test_get_new_files_after_adding_files(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 12)

        create_versioned_files(self.testInst, start, stop, freq='100min',
                               use_doy=False,
                               root_fname=self.root_fname)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        new_files = self.testInst.files.get_new()
        assert (np.all(new_files.index == dates))

    def test_get_new_files_after_adding_files_and_adding_file(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 12)

        create_versioned_files(self.testInst, start, stop, freq='100min',
                               use_doy=False,
                               root_fname=self.root_fname)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        new_files = self.testInst.files.get_new()

        start = pysat.datetime(2008, 1, 15)
        stop = pysat.datetime(2008, 1, 18)

        create_versioned_files(self.testInst, start, stop, freq='100min',
                               use_doy=False,
                               root_fname=self.root_fname)
        dates2 = pysat.utils.time.create_date_range(start, stop, freq='100min')
        new_files2 = self.testInst.files.get_new()
        assert (np.all(new_files.index == dates) &
                np.all(new_files2.index == dates2))

    def test_get_new_files_after_deleting_files_and_adding_files(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 12)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        # remove files, same number as will be added
        to_be_removed = len(dates)
        for the_file in os.listdir(self.testInst.files.data_path):
            if (the_file[0:13] == 'pysat_testing') & \
                    (the_file[-19:] == '.pysat_testing_file'):
                file_path = os.path.join(self.testInst.files.data_path,
                                         the_file)
                if os.path.isfile(file_path) & (to_be_removed > 0):
                    to_be_removed -= 1
                    # Remove all versions of the file
                    # otherwise, previous versions will look like new files
                    pattern = '_'.join(file_path.split('_')[0:7]) + \
                        '*.pysat_testing_file'
                    map(os.unlink, glob.glob(pattern))
                    # os.unlink(file_path)
        # add new files
        create_versioned_files(self.testInst, start, stop, freq='100min',
                               use_doy=False, root_fname=self.root_fname)
        # get new files
        new_files = self.testInst.files.get_new()

        assert (np.all(new_files.index == dates))

    def test_files_non_standard_pysat_directory(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 15)
        dates = pysat.utils.time.create_date_range(start, stop, freq='100min')
        pysat.instruments.pysat_testing.list_files = list_versioned_files
        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             sat_id='hello',
                             directory_format='pysat_testing_{tag}_{sat_id}',
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)
        # add new files
        create_dir(self.testInst)
        remove_files(self.testInst)
        create_versioned_files(self.testInst, start, stop, freq='100min',
                               use_doy=False, root_fname=self.root_fname)

        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             sat_id='hello',
                             directory_format='pysat_testing_{tag}_{sat_id}',
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)

        # get new files
        new_files = self.testInst.files.get_new()
        assert (np.all(self.testInst.files.files.index == dates) &
                np.all(new_files.index == dates))

    def test_files_non_standard_file_format_template(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 15)
        dates = pysat.utils.time.create_date_range(start, stop, freq='1D')

        # clear out old files, create new ones
        remove_files(self.testInst)
        create_versioned_files(self.testInst, start, stop, freq='1D',
                               use_doy=False,
                               root_fname=''.join(('pysat_testing_unique_',
                                                   '{version:02d}_',
                                                   '{revision:03d}_{year:04d}',
                                                   '_g_{day:03d}_st.pysat_',
                                                   'testing_file')))

        pysat.instruments.pysat_testing.list_files = list_versioned_files
        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             file_format=''.join(('pysat_testing_unique_',
                                                  '{version:02d}_',
                                                  '{revision:03d}_{year:04d}_',
                                                  'g_{day:03d}_st.pysat_',
                                                  'testing_file')),
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)
        assert (np.all(self.testInst.files.files.index == dates))

    def test_files_when_duplicates_forced(self):
        # create new files and make sure that new files are captured
        start = pysat.datetime(2008, 1, 11)
        stop = pysat.datetime(2008, 1, 15)
        dates = pysat.utils.time.create_date_range(start, stop, freq='1D')

        # clear out old files, create new ones
        remove_files(self.testInst)
        create_versioned_files(self.testInst, start, stop, freq='1D',
                               use_doy=False,
                               root_fname=''.join(('pysat_testing_unique_',
                                                   '{version:02d}_',
                                                   '{revision:03d}_{year:04d}',
                                                   '_g_{day:03d}_st.pysat_',
                                                   'testing_file')))

        pysat.instruments.pysat_testing.list_files = list_files
        self.testInst = \
            pysat.Instrument(inst_module=pysat.instruments.pysat_testing,
                             clean_level='clean',
                             file_format=''.join(('pysat_testing_unique_??_',
                                                  '???_{year:04d}_g_{day:03d}',
                                                  '_st.pysat_testing_file')),
                             update_files=True,
                             temporary_file_list=self.temporary_file_list)
        assert (np.all(self.testInst.files.files.index == dates))


class TestInstrumentWithVersionedFilesNoFileListStorage(TestInstrumentWithVersionedFiles):

    temporary_file_list = True
